# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import inotify.adapters
import inotify.constants

from trackma import utils
from trackma.tracker import inotifyBase


class inotifyTracker(inotifyBase.inotifyBase):
    name = 'Tracker (inotify)'

    def observe(self, config, watch_dirs):
        # Note that this lib uses bytestrings for filenames and paths.
        self.msg.info('Using inotify.')

        self.msg.debug('Checking if there are open players...')
        opened = self._proc_poll()
        if opened:
            self._proc_open(*opened)

        self.msg.debug('Watching the following paths: ' + ','.join(watch_dirs))

        mask = (inotify.constants.IN_OPEN
                | inotify.constants.IN_CLOSE
                | inotify.constants.IN_CREATE
                | inotify.constants.IN_MOVE
                | inotify.constants.IN_DELETE)

        i = inotify.adapters.InotifyTrees(list(watch_dirs), mask=mask)

        try:
            for event in i.event_gen():
                if not self.active:
                    return

                if event is not None:
                    # With inotifyx impl., only the event type was used,
                    # such that it only served to poll lsof when an
                    # open or close event was received.
                    (header, types, path, filename) = event
                    if 'IN_ISDIR' not in types:
                        # If the file is gone, we remove from library
                        if ('IN_MOVED_FROM' in types
                                or 'IN_DELETE' in types):
                            self._emit_signal('removed', path, filename)
                        # Otherwise we attempt to add it to library
                        # Would check for IN_MOVED_TO or IN_CREATE but no need
                        else:
                            self._emit_signal('detected', path, filename)

                        if 'IN_OPEN' in types:
                            self._proc_open(path, filename)
                        elif ('IN_CLOSE_NOWRITE' in types
                              or 'IN_CLOSE_WRITE' in types):
                            self._proc_close(path, filename)
                elif self.last_state != utils.Tracker.NOVIDEO and not self.last_updated:
                    # Default blocking duration is 1 second
                    # This will count down like inotifyx impl. did
                    self.update_show_if_needed(
                        self.last_state, self.last_show_tuple)
        finally:
            self.msg.info('Tracker has stopped.')
            # inotify resource is cleaned-up automatically



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt6.QtWidgets import QDialog, QDialogButtonBox, QVBoxLayout

from trackma.ui.qt.widgets import DetailsWidget


class DetailsDialog(QDialog):
    def __init__(self, parent, worker, show):
        QDialog.__init__(self, parent)
        self.setMinimumSize(530, 550)
        self.setWindowTitle('Details')
        self.worker = worker

        main_layout = QVBoxLayout()
        details = DetailsWidget(self, worker)

        bottom_buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
        bottom_buttons.setCenterButtons(True)
        bottom_buttons.rejected.connect(self.close)

        main_layout.addWidget(details)
        main_layout.addWidget(bottom_buttons)

        self.setLayout(main_layout)
        details.load(show)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re
import os
import anitopy
from decimal import Decimal


class AnitopyWrapper():
    """
    Wrapper class around Anitopy, the anime filename parser.
    Exists mainly for compatibility reasons, but also to work around
    some edge cases that Anitopy cannot solve.
    """

    def __init__(self, msg, file_name):
        self.msg = msg.with_classname('Parser')
        self.original_file_name = file_name

        file_name = self.__preProcessFileName(file_name)
        file_name = self.__trimFileName(file_name)
        self.file_name = file_name

        try:
            self.msg.debug(f"Parsing {file_name}")
            data = anitopy.parse(file_name)
        except Exception:
            # If Anitopy crashes while parsing a filename, print the traceback
            # instead of crashing Trackma altogether.
            import traceback
            traceback.print_exc()
            data = {}
            return

        self.episode_number = self.__extractEpisodeNumber(data)
        self.anime_title = self.__extractAnimeTitle(data)

    def getName(self):
        # Returns the anime title
        return self.anime_title

    def getEpisode(self):
        # Returns the first/only episode number
        if self.episode_number is None:
            return 1

        if type(self.episode_number) is list:
            return int(self.episode_number[-1])
        else:
            return int(self.episode_number)

    def getEpisodeNumbers(self, force_numbers=False):
        # Returns the episode range as a tuple
        (ep_start, ep_end) = (None, None)

        if self.episode_number:
            try:
                if isinstance(self.episode_number, list):
                    ep_start = Decimal(self.episode_number[0])
                    ep_end = Decimal(self.episode_number[-1])
                else:
                    ep_start = Decimal(self.episode_number)
            except ArithmeticError:
                self.msg.warn("Unable to parse episode number '{}' of: {}"
                              .format(self.episode_number, self.original_file_name))

        if force_numbers:
            if ep_start is None:
                ep_start = 1
            if ep_end is None:
                ep_end = ep_start
            (ep_start, ep_end) = (int(ep_start), int(ep_end))

        return (ep_start, ep_end)

    @staticmethod
    def __preProcessFileName(file_name):
        # Make some adjustments to the file name to increase parsing accuracy

        # If full path is provided to Anitopy, all [brackets with contents]
        # adjacent to the path separators should be moved to the VERY beginning.
        # Or Anitopy won't be able to parse them out.
        file_name = os.path.sep + file_name
        for m in re.finditer(
                r'(?<={0})\[.*?\]|\[.*?\](?={0})'.format(os.path.sep),
                file_name):
            file_name = (m.group() + file_name[:m.start()]
                              + file_name[m.end():])

        # Remove all the path separators (except the last one, we'll need it later)
        parts = file_name.split(os.path.sep)
        file_name = ' '.join(parts)

        # Anitopy can parse S01E01 properly, but not S01OVA01, S01S01, S01NCOP01 etc.
        # So we'll need to break things down for the parser.
        m = re.search(r'S(?P<season>[0-9]+)(?P<type>[A-Za-z]+)(?P<episode>[0-9]+)', file_name)
        if m:
            groups = m.groupdict()
            # 'S01E01' -> 'Season 01 - 01'
            if groups['type'] == 'E':
                groups['type'] = ''
            # 'S01S01' -> 'Season 01 Specials - 01'
            if groups['type'] == 'S':
                groups['type'] = 'Specials'
            # for all other cases:
            # 'S01{type}01' -> 'Season 01 {type} - 01'
            file_name = (
                file_name[:m.start()] + 'Season ' + groups['season']
                + ' ' + groups['type'] + ' - ' + groups['episode']
                + file_name[m.end():])

        return file_name

    @staticmethod
    def __trimFileName(file_name):
        # If the same title appears in the parent directory and the file name,
        # we might want to remove the duplicate one.
        # Otherwise, Anitopy would concatenate them together.
        try:
            # Temporarily replace all punctuations with spaces
            temp = re.sub(r'[^\w\s{0}\(\)\{{\}}\[\]]'.format(os.path.sep),
                          r' ', file_name, flags=re.ASCII)
            # Search and remove the longest duplicate
            m = max(
                [x for x in re.finditer(
                    r'(\b.{{3,}}\b)(?=.*?{0}.*?(?P<DUP>\1))'.format(os.path.sep),
                    temp, flags=re.IGNORECASE)],
                key = lambda y: y.end() - y.start()
            )
            if m:
                file_name = file_name[:m.start('DUP')] + ' ' + file_name[m.end('DUP'):]
        except ValueError:
            pass

        # Remove the remaining path separator(s)
        file_name = file_name.replace(os.path.sep, ' ')
        # Remove empty ( ) brackets
        file_name = re.sub(r'[\[\{\(][\s\._]*[\)\}\]]', r'', file_name)
        # Trim unnecessary     spaces
        file_name = re.sub(r'\s{2,}', r' ', file_name.strip())

        return file_name

    @staticmethod
    def __extractAnimeTitle(data):
        # Deal with anime title related stuff that Anitopy left out
        if 'anime_title' not in data:
            return None
        anime_title = data['anime_title']

        # Append anime season to the title (if needed)
        anime_season = data.get('anime_season')
        if anime_season:
            if isinstance(anime_season, list):
                anime_season = anime_season[0]
            if int(anime_season) > 1:
                anime_title += ' Season ' + anime_season

        # Solve 'Season X Part Y' cases
        if 'episode_title' in data:
            m = re.search(r'^Part [2-9]\b', data['episode_title'], flags=re.IGNORECASE)
            if m:
                anime_title += ' ' + m.group(0)

        # Append anime type to the title (if needed)
        anitype_invalid = ('OP', 'NCOP', 'OPENING', 'ED', 'NCED', 'ENDING', 'PV', 'PREVIEW')
        anitype_specials = ('OAD', 'OAV', 'ONA', 'OVA', 'SPECIAL', 'SPECIALS')
        anitype = data.get('anime_type')
        if anitype:
            if not isinstance(anitype, list):
                anitype = [anitype]
            for t in anitype:
                # Ignore non-episodes such as openings, endings, previews etc.
                if t.upper() in anitype_invalid:
                    return None
                if t not in anime_title and t.upper() in anitype_specials:
                    anime_title += ' ' + t
        else:
            # Fix anime type being detected as episode title
            if 'episode_title' in data:
                for t in anitype_specials:
                    m = re.search(r'{0}\b'.format(re.escape(t)), data['episode_title'],
                                  flags=re.IGNORECASE)
                    if m:
                        anime_title += ' ' + m.group(0)

        # Append anime year to the title (if needed)
        anime_year = data.get('anime_year')
        if anime_year and anime_year not in anime_title:
            anime_title += ' (' + anime_year + ')'

        return anime_title

    @staticmethod
    def __extractEpisodeNumber(data):
        # Deal with episode related stuff that Anitopy left out
        if 'episode_number' not in data:
            return
        episode_number = data['episode_number']

        # Handle cases like: "[Judas] Naruto - S05E01 (186).mkv"
        # Anitopy should detect the consecutive episode number (186) properly.
        # Just set that as the original episode number, and remove the season value.
        if 'episode_number_alt' in data:
            episode_number = data['episode_number_alt']
            del data['episode_number_alt']
            if 'anime_season' in data:
                del data['anime_season']

        # Unfortunately, we can't have episode numbers like 1A, 1B, 1C etc.
        if isinstance(data['episode_number'], str):
            episode_number = re.sub(r'ABCabc', r'', episode_number)

        return episode_number



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import pyinotify

from trackma import utils
from trackma.tracker import inotifyBase


class pyinotifyTracker(inotifyBase.inotifyBase):
    name = 'Tracker (pyinotify)'

    def observe(self, config, watch_dirs):
        self.msg.info('Using pyinotify.')

        self.msg.debug('Checking if there are open players...')
        opened = self._proc_poll()
        if opened:
            self._proc_open(*opened)

        wm = pyinotify.WatchManager()  # Watch Manager
        mask = (pyinotify.IN_OPEN  # pylint: disable=no-member
                | pyinotify.IN_CLOSE_NOWRITE  # pylint: disable=no-member
                | pyinotify.IN_CLOSE_WRITE  # pylint: disable=no-member
                | pyinotify.IN_CREATE  # pylint: disable=no-member
                | pyinotify.IN_MOVED_FROM  # pylint: disable=no-member
                | pyinotify.IN_MOVED_TO  # pylint: disable=no-member
                | pyinotify.IN_DELETE)  # pylint: disable=no-member

        class EventHandler(pyinotify.ProcessEvent):
            def my_init(self, parent=None):
                self.parent = parent

            def process_IN_OPEN(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._proc_open(event.path, event.name)

            def process_IN_CLOSE_NOWRITE(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._proc_close(event.path, event.name)

            def process_IN_CLOSE_WRITE(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._proc_close(event.path, event.name)

            def process_IN_CREATE(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._emit_signal(
                        'detected', event.path, event.name)

            def process_IN_MOVED_TO(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._emit_signal(
                        'detected', event.path, event.name)

            def process_IN_MOVED_FROM(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._emit_signal('removed', event.path, event.name)

            def process_IN_DELETE(self, event):
                if not event.mask & pyinotify.IN_ISDIR:  # pylint: disable=no-member
                    self.parent._emit_signal('removed', event.path, event.name)

        handler = EventHandler(parent=self)
        notifier = pyinotify.Notifier(wm, handler)
        for path in watch_dirs:
            self.msg.debug('Watching directory {}'.format(path))
            wm.add_watch(path, mask, rec=True, auto_add=True)

        try:
            # notifier.loop()
            timeout = None
            while self.active:
                if notifier.check_events(timeout):
                    # Check again to avoid notifying while inactive
                    if not self.active:
                        return

                    notifier.read_events()
                    notifier.process_events()
                    if self.last_state == utils.Tracker.NOVIDEO or self.last_updated:
                        timeout = None  # Block indefinitely
                    else:
                        timeout = 1000  # Check each second for counting
                elif self.active:
                    self.msg.debug("Sending last state {} {}".format(
                        self.last_state, self.last_show_tuple))
                    self.update_show_if_needed(
                        self.last_state, self.last_show_tuple)
        finally:
            notifier.stop()
            self.msg.info('Tracker has stopped.')



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import html
import os
import threading

from gi.repository import GLib, GObject, Gdk, Gtk

from trackma import messenger
from trackma import utils
from trackma.ui.gtk import gtk_dir
from trackma.ui.gtk.imagebox import ImageBox
from trackma.ui.gtk.showeventtype import ShowEventType
from trackma.ui.gtk.showtreeview import ShowListFilter, ShowListStore, ShowTreeView


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/mainview.ui'))
class MainView(Gtk.Box):

    __gtype_name__ = 'MainView'

    __gsignals__ = {
        'error': (GObject.SignalFlags.RUN_FIRST, None,
                  (str, )),
        'success': (GObject.SignalFlags.RUN_CLEANUP, None,
                    ()),
        'error-fatal': (GObject.SignalFlags.RUN_FIRST, None,
                        (str,)),
        'show-action': (GObject.SignalFlags.RUN_FIRST, None,
                        (int, object)),
    }

    image_container_box = Gtk.Template.Child()
    top_box = Gtk.Template.Child()
    show_title = Gtk.Template.Child()
    api_icon = Gtk.Template.Child()
    api_user = Gtk.Template.Child()
    btn_episode_remove = Gtk.Template.Child()
    btn_episode_show_entry = Gtk.Template.Child()
    entry_popover = Gtk.Template.Child()
    entry_episode = Gtk.Template.Child()
    entry_done = Gtk.Template.Child()
    btn_episode_add = Gtk.Template.Child()
    btn_play_next = Gtk.Template.Child()
    spinbtn_score = Gtk.Template.Child()
    btn_score_set = Gtk.Template.Child()
    statusbox = Gtk.Template.Child()
    statusmodel = Gtk.Template.Child()
    notebook = Gtk.Template.Child()

    def __init__(self, config, debug=False):
        Gtk.Box.__init__(self)
        self.init_template()

        self._configfile = utils.to_config_path('ui-Gtk.json')
        self._config = config
        self._engine = None
        self._account = None
        self._list = ShowListStore(colors=config['colors'])
        self._debug = debug

        self._image_thread = None
        self._current_page = None
        self.statusbox_handler = None
        self.notebook_switch_handler = None
        self._hovering_over_tabs = None
        self._pages = {}
        self._page_handler_ids = {}

        self._init_widgets()
        self._init_signals()

    def load_engine_account(self, engine, account):
        self._engine = engine
        self._account = account

        self._engine_start()
        self._init_signals_engine()

    def load_account_mediatype(self, account, mediatype, extern_widget):
        if account:
            self._account = account

        self._engine_reload(account, mediatype, extern_widget)

    def _init_widgets(self):
        self.image_box = ImageBox(100, 150)
        self.image_box.show()
        self.image_container_box.pack_start(self.image_box, False, False, 0)

        self.notebook.set_scrollable(True)
        self.notebook.add_events(Gdk.EventMask.SCROLL_MASK |
                                 Gdk.EventMask.SMOOTH_SCROLL_MASK)
        self.notebook.connect('scroll-event', self._notebook_handle_events)
        self.notebook.connect('enter-notify-event',
                              self._notebook_handle_events)
        self.notebook.connect('leave-notify-event',
                              self._notebook_handle_events)

        self.statusbar = Gtk.Statusbar()
        self.statusbar.push(0, 'Trackma GTK ' + utils.VERSION)
        self.statusbar.show()
        self.pack_start(self.statusbar, False, False, 0)

    def _init_signals(self):
        self.btn_episode_remove.connect(
            "clicked", self._on_btn_episode_remove_clicked)
        self.btn_episode_show_entry.connect(
            "clicked", self._show_episode_entry)
        self.entry_episode.connect("activate", self._on_entry_episode_activate)
        self.entry_done.connect("clicked", self._on_entry_episode_activate)
        self.entry_popover.connect("focus-out-event", self._hide_episode_entry)
        self.btn_episode_add.connect(
            "clicked", self._on_btn_episode_add_clicked)
        self.btn_play_next.connect(
            "clicked", self._on_btn_play_next_clicked, True)
        self.spinbtn_score.connect("activate", self._on_spinbtn_score_activate)
        self.btn_score_set.connect("clicked", self._on_spinbtn_score_activate)
        self.statusbox_handler = self.statusbox.connect(
            "changed", self._on_statusbox_changed)
        self.notebook_switch_handler = self.notebook.connect(
            "switch-page", self._on_switch_notebook_page)

    def _init_signals_engine(self):
        self._engine.connect_signal(
            'episode_changed', self._on_changed_show_idle)
        self._engine.connect_signal(
            'score_changed', self._on_changed_show_idle)
        self._engine.connect_signal(
            'status_changed', self._on_changed_show_status_idle)
        self._engine.connect_signal('playing', self._on_playing_show_idle)
        self._engine.connect_signal(
            'show_added', self._on_changed_show_status_idle)
        self._engine.connect_signal(
            'show_deleted', self._on_changed_show_status_idle)
        self._engine.connect_signal(
            'prompt_for_update', self._on_prompt_update_next_idle)

    def _engine_start(self):
        threading.Thread(target=self._engine_start_task).start()

    def _engine_start_task(self):
        if self._engine.loaded:
            return

        try:
            self._engine.start()
        except utils.TrackmaFatal as e:
            self.emit('error-fatal', e)
            return

        GLib.idle_add(self._update_widgets)

    def _engine_reload(self, account, mediatype, extern_widget):
        self.set_buttons_sensitive(False)
        threading.Thread(target=self._engine_reload_task,
                         args=[account, mediatype, extern_widget]).start()

    def _engine_reload_task(self, account, mediatype, extern_widget):
        try:
            self._engine.reload(account, mediatype)
        except utils.TrackmaError as e:
            self.emit('error', e)
        except utils.TrackmaFatal as e:
            self.emit('error-fatal', e)
            return

        GLib.idle_add(self._update_widgets, extern_widget)

    def _update_widgets(self, extern_widget=None):
        self.statusbox.handler_block(self.statusbox_handler)
        self._reset_widgets()
        self._create_notebook_pages()
        self._set_score_ranges()
        self.populate_all_pages()
        self._populate_statusbox()
        self.statusbox.handler_unblock(self.statusbox_handler)
        if extern_widget is not None:
            extern_widget.set_subtitle(self._engine.api_info['name'] + " (" +
                                       self._engine.api_info['mediatype'] + ")")

        self.set_status_idle("Ready.")
        self.emit('success')
        self.set_buttons_sensitive_idle(True)

    def _reset_widgets(self):
        self.show_title.set_text('<span size="14000"><b>Trackma</b></span>')
        self.show_title.set_use_markup(True)
        self.image_box.reset()

        current_api = utils.available_libs[self._account['api']]
        api_iconfile = current_api[1]

        self.api_icon.set_from_file(api_iconfile)

        self.api_user.set_text("%s" % self._engine.get_userconfig('username'))

        can_play = self._engine.mediainfo['can_play']
        can_update = self._engine.mediainfo['can_update']

        self.btn_play_next.set_sensitive(can_play)
        self.btn_episode_show_entry.set_sensitive(can_update)
        self.entry_episode.set_sensitive(can_update)
        self.entry_done.set_sensitive(can_update)
        self.btn_episode_add.set_sensitive(can_update)

    def _notebook_handle_events(self, widget, event):
        if event.type in (Gdk.EventType.LEAVE_NOTIFY, Gdk.EventType.ENTER_NOTIFY):
            self._hovering_over_tabs = event.type == Gdk.EventType.ENTER_NOTIFY
        elif self._hovering_over_tabs and event.type == Gdk.EventType.SCROLL:
            page = self.notebook.get_current_page()
            npage = self.notebook.get_n_pages() - 1
            if event.delta_y < 0 and page > 0:
                self.notebook.prev_page()
            elif event.delta_y > 0 and page < npage:
                self.notebook.next_page()
            return event.delta_y < 0 and page > 0 or event.delta_y > 0 and page < npage
        return False

    def _create_notebook_pages(self):
        statuses_nums = self._engine.mediainfo['statuses'].copy()
        statuses_names = self._engine.mediainfo['statuses_dict'].copy()
        statuses_nums.append(None)
        statuses_names[None] = 'All'
        self.notebook.handler_block(self.notebook_switch_handler)
        # Clear notebook
        for i in range(self.notebook.get_n_pages()):
            self.notebook.remove_page(-1)

        self._pages = {}
        self._page_handler_ids = {}

        # Insert pages
        for status in statuses_nums:
            page_title = Gtk.Label(label=statuses_names[status])
            self._pages[status] = NotebookPage(self._engine,
                                               self.notebook.get_n_pages(),
                                               status,
                                               self._config,
                                               self._list,
                                               page_title)

            self._page_handler_ids[status] = []
            self._page_handler_ids[status].append(
                self._pages[status].connect('show-selected', self._on_show_selected))
            self._page_handler_ids[status].append(
                self._pages[status].connect('show-action', self._on_show_action))
            self._page_handler_ids[status].append(
                self._pages[status].connect('column-toggled', self._on_column_toggled))
            self.notebook.append_page(self._pages[status],
                                      page_title)

        self.notebook.handler_unblock(self.notebook_switch_handler)
        self.notebook.show_all()

    def populate_all_pages(self):
        for status in self._pages:
            self._block_handlers_for_status(status)
            tree_view = self._pages[status].show_tree_view
            tree_view.freeze_child_notify()

        self._list.clear()
        library = self._engine.library()
        for show in self._engine.get_list():
            self._list.append(show,
                              self._engine.altname(show['id']),
                              library.get(show['id']))

        self._list.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        for status in self._pages:
            tree_view = self._pages[status].show_tree_view
            tree_view.thaw_child_notify()
            self._unblock_handlers_for_status(status)

    def _block_handlers_for_status(self, status):
        for handler_id in self._page_handler_ids[status]:
            self._pages[status].handler_block(handler_id)

    def _unblock_handlers_for_status(self, status):
        for handler_id in self._page_handler_ids[status]:
            self._pages[status].handler_unblock(handler_id)

    def _populate_statusbox(self):
        statuses_nums = self._engine.mediainfo['statuses']
        statuses_names = self._engine.mediainfo['statuses_dict']

        self.statusmodel.clear()
        for status in statuses_nums:
            self.statusmodel.append([str(status), statuses_names[status]])
        self.statusbox.set_model(self.statusmodel)
        self.statusbox.show_all()

    def _set_score_ranges(self):
        score_decimal_places = 0
        if isinstance(self._engine.mediainfo['score_step'], float):
            score_decimal_places = len(
                str(self._engine.mediainfo['score_step']).split('.')[1])

        self.spinbtn_score.set_value(0)
        self.spinbtn_score.set_digits(score_decimal_places)
        self.spinbtn_score.set_range(0, self._engine.mediainfo['score_max'])
        self.spinbtn_score.get_adjustment().set_step_increment(
            self._engine.mediainfo['score_step'])

        for view in self._pages.values():
            view.decimals = score_decimal_places

    def set_status_idle(self, msg):
        # Thread safe
        GLib.idle_add(self._set_status, msg)

    def _set_status(self, msg):
        print(msg)
        self.statusbar.push(0, msg)

    def set_buttons_sensitive_idle(self, boolean):
        # Thread safe
        GLib.idle_add(self.set_buttons_sensitive, boolean)

    def set_buttons_sensitive(self, boolean, lists_too=True):
        if lists_too:
            self.notebook.set_sensitive(boolean)

        if (self._current_page and self._current_page.selected_show) or not boolean:
            if self._engine.mediainfo['can_play']:
                self.btn_play_next.set_sensitive(boolean)

            if self._engine.mediainfo['can_update']:
                self.btn_episode_show_entry.set_sensitive(boolean)
                self.entry_episode.set_sensitive(boolean)
                self.entry_done.set_sensitive(boolean)
                self.btn_episode_add.set_sensitive(boolean)
                self.btn_episode_remove.set_sensitive(boolean)

            self.btn_score_set.set_sensitive(boolean)
            self.spinbtn_score.set_sensitive(boolean)
            self.statusbox.set_sensitive(boolean)

    def _on_btn_episode_remove_clicked(self, widget):
        self.emit('show-action',
                  ShowEventType.EPISODE_REMOVE,
                  (self._current_page.selected_show,))

    def _show_episode_entry(self, widget):
        self.entry_popover.set_relative_to(widget)
        self.entry_popover.set_position(Gtk.PositionType.BOTTOM)
        self.entry_episode.set_text(self.btn_episode_show_entry.get_label())
        self.entry_popover.show()
        self.entry_episode.grab_focus()

    def _on_entry_episode_activate(self, widget):
        try:
            episode = int(self.entry_episode.get_text())
            self.emit('show-action',
                      ShowEventType.EPISODE_SET,
                      (self._current_page.selected_show, episode))
            self._hide_episode_entry()
        except ValueError:
            pass

    def _hide_episode_entry(self, *args):
        self.entry_popover.hide()

    def _on_btn_episode_add_clicked(self, widget):
        self.emit('show-action',
                  ShowEventType.EPISODE_ADD,
                  (self._current_page.selected_show,))

    def _on_btn_play_next_clicked(self, widget, playnext, ep=None):
        self.emit('show-action',
                  ShowEventType.PLAY_NEXT,
                  (self._current_page.selected_show,))

    def _on_spinbtn_score_activate(self, widget):
        score = round(self.spinbtn_score.get_value(), self.spinbtn_score.get_digits())
        self.emit('show-action',
                  ShowEventType.SET_SCORE,
                  (self._current_page.selected_show, score))

    def _on_statusbox_changed(self, widget):
        statusiter = self.statusbox.get_active_iter()
        status = self.statusmodel.get(statusiter, 0)[0]
        self.emit('show-action',
                  ShowEventType.SET_STATUS,
                  (self._current_page.selected_show, status))

    def message_handler(self, classname, msgtype, msg):
        # Thread safe
        if msgtype == messenger.TYPE_WARN:
            self.set_status_idle("%s warning: %s" % (classname, msg))
        elif msgtype != messenger.TYPE_DEBUG:
            self.set_status_idle("%s: %s" % (classname, msg))
        elif self._debug:
            print('[D] {}: {}'.format(classname, msg))

    def _on_changed_show_idle(self, show):
        GLib.idle_add(self._update_show, show)

    def _update_show(self, show):
        status = show['my_status']
        self._list.update(show)
        if self._current_page and show['id'] == self._current_page.selected_show:
            self.btn_episode_show_entry.set_label(str(show['my_progress']))
            self.spinbtn_score.set_value(show['my_score'])

    def change_show_title_idle(self, show, altname):
        GLib.idle_add(self._update_show_title, show, altname)

    def _update_show_title(self, show, altname):
        self._list.update_title(show, altname)

    def _on_changed_show_status_idle(self, show, old_status=None):
        GLib.idle_add(self._update_show_status, show, old_status)

    def _update_show_status(self, show, old_status):
        # Rebuild lists
        status = show['my_status']
        try:
            self._engine.get_show_info(showid=show['id'])
            self._list.update_or_append(show)
        except utils.EngineError:
            self._list.remove(show)
        pagenumber = self._pages[status].pagenumber
        current_page = self.notebook.get_current_page()
        if current_page != pagenumber:
            if current_page != self._pages[None].pagenumber:
                self.notebook.set_current_page(pagenumber)
                self._pages[status].show_tree_view.select(show)
            else:
                self._pages[None].show_tree_view.select(show)

    def _on_playing_show_idle(self, show, is_playing, episode):
        GLib.idle_add(self._set_show_playing, show, is_playing, episode)

    def _set_show_playing(self, show, is_playing, episode):
        self._list.playing(show, is_playing)

    def _on_prompt_update_next_idle(self, show, played_ep):
        GLib.idle_add(self._prompt_update_next, show, played_ep)

    def _prompt_update_next(self, show, played_ep):
        dialog = Gtk.MessageDialog(self.get_toplevel(),
                                   Gtk.DialogFlags.MODAL,
                                   Gtk.MessageType.QUESTION,
                                   Gtk.ButtonsType.YES_NO,
                                   "Update %s to episode %d?" % (show['title'], played_ep))
        dialog.show_all()
        dialog.connect(
            "response", self._on_response_update_next, show, played_ep)

    def _on_response_update_next(self, widget, response, show, played_ep):
        widget.destroy()
        if response == Gtk.ResponseType.YES:
            self.emit('show-action',
                      ShowEventType.EPISODE_SET,
                      (show['id'], played_ep))

    def _on_switch_notebook_page(self, notebook, page, page_num):
        self._current_page = page
        self._update_widgets_for_selected_show()

    def _on_show_selected(self, page, selected_show):
        self._update_widgets_for_selected_show()

    def _update_widgets_for_selected_show(self):
        if not (self._current_page and self._current_page.selected_show):
            self.set_buttons_sensitive(False, lists_too=False)
            return

        self.set_buttons_sensitive(True, lists_too=False)
        try:
            show = self._engine.get_show_info(self._current_page.selected_show)
        except utils.EngineError:
            self._list.remove(show_id=self._current_page.selected_show)
            self.set_buttons_sensitive(False, lists_too=False)
            return

        # Block handlers
        self.statusbox.handler_block(self.statusbox_handler)

        if self._image_thread is not None:
            self._image_thread.cancel()

        self.show_title.set_text(
            '<span size="14000"><b>{0}</b></span>'.format(html.escape(show['title'])))
        self.show_title.set_use_markup(True)

        # Episode selector
        self.btn_episode_show_entry.set_label(str(show['my_progress']))

        # Status selector
        for i in self.statusmodel:
            if str(i[0]) == str(show['my_status']):
                self.statusbox.set_active_iter(i.iter)
                break

        # Score selector
        self.spinbtn_score.set_value(show['my_score'])

        # Image
        if show.get('image_thumb') or show.get('image'):
            utils.make_dir(utils.to_cache_path())
            filename = utils.to_cache_path(
                "%s_%s_%s.jpg" % (self._engine.api_info['shortname'],
                                  self._engine.api_info['mediatype'],
                                  show['id']))

            if os.path.isfile(filename):
                self.image_box.set_image(filename)
            else:
                self.image_box.set_image_remote(show.get('image_thumb') or show['image'],
                                                filename)
        else:
            self.image_box.set_text('No Image')

        # Unblock handlers
        self.statusbox.handler_unblock(self.statusbox_handler)

    def _on_show_action(self, page, event_type, data):
        self.emit('show-action', event_type, data)

    def get_current_status(self):
        print(self._engine.mediainfo['statuses'])
        return self._current_page.status if self._current_page.status is not None else self._engine.mediainfo['statuses'][-1]

    def get_selected_show(self):
        if not self._current_page:
            return None

        return self._current_page.selected_show

    def _on_column_toggled(self, page, column_name, visible):
        if visible:
            # Make column visible
            self._config['visible_columns'].append(column_name)
        else:
            # Make column invisible
            if len(self._config['visible_columns']) <= 1:
                return  # There should be at least 1 column visible

            self._config['visible_columns'].remove(column_name)

        for page in self._pages.values():
            page.set_column_visible(column_name, visible)

        utils.save_config(self._config, self._configfile)


class NotebookPage(Gtk.ScrolledWindow):
    __gtype_name__ = 'NotebookPage'

    __gsignals__ = {
        'show-selected': (GObject.SignalFlags.RUN_FIRST, None,
                          (int, )),
        'show-action': (GObject.SignalFlags.RUN_FIRST, None,
                        (int, object)),
        'column-toggled': (GObject.SignalFlags.RUN_FIRST, None,
                           (str, bool)),
    }

    def __init__(self, engine, page_num, status, config, _list=None, title=None):
        super().__init__()
        self._engine = engine
        self._page_number = page_num
        self._status = status
        self._selected_show = 0
        self._list = _list
        self._title = title
        self._title_text = self._engine.mediainfo['statuses_dict'][status] if status in self._engine.mediainfo['statuses_dict'].keys(
        ) else 'All'
        self._init_widgets(page_num, status, config)

    def _init_widgets(self, page_num, status, config):
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.set_size_request(550, 300)
        self.set_border_width(5)

        self._show_tree_view = ShowTreeView(
            config['colors'],
            config['visible_columns'],
            config['episodebar_style'])
        self._show_tree_view.set_model(
            Gtk.TreeModelSort(
                model=ShowListFilter(
                    status=self.status,
                    child_model=self._list
                )
            )
        )
        self._title.set_text('%s (%d)' % (
            self._title_text,
            len(self._show_tree_view.props.model)
        ))
        self._show_tree_view.get_selection().connect(
            "changed", self._on_selection_changed)
        self._show_tree_view.connect("row-activated", self._on_row_activated)
        self._show_tree_view.connect("column-toggled", self._on_column_toggled)
        self._show_tree_view.connect(
            "button-press-event", self._on_show_context_menu)
        self._show_tree_view.get_model().connect("row-inserted", self._update_title)
        self._show_tree_view.get_model().connect("row-deleted", self._update_title)

        self.add(self._show_tree_view)

    def set_column_visible(self, column_name, visible):
        self._show_tree_view.cols[column_name].set_visible(visible)

    def _update_title(self, model, *args):
        self._title.set_text('%s (%d)' % (
            self._title_text,
            len(model)
        ))

    @property
    def decimals(self):
        return self._list.decimals

    @decimals.setter
    def decimals(self, decimals):
        self._list.decimals = decimals

    @property
    def status(self):
        return self._status

    @property
    def pagenumber(self):
        return self._page_number

    @property
    def selected_show(self):
        return self._selected_show

    @property
    def show_tree_view(self):
        return self._show_tree_view

    def _on_selection_changed(self, selection):
        (tree_model, tree_iter) = selection.get_selected()
        if not tree_iter:
            self._selected_show = 0
            return

        try:
            self._selected_show = int(tree_model.get(tree_iter, 0)[0])
        except ValueError:
            self._selected_show = tree_model.get(tree_iter, 0)[0]

        self.emit('show-selected', self._selected_show)

    def _on_row_activated(self, tree_view, path, column):
        self.emit('show-action', ShowEventType.DETAILS, (self.selected_show,))

    def _on_column_toggled(self, tree_view, column_name, visible):
        self.emit('column-toggled', column_name, visible)

    def _on_show_context_menu(self, tree_view, event):
        x = int(event.x)
        y = int(event.y)
        pthinfo = tree_view.get_path_at_pos(x, y)

        if (event.type == Gdk.EventType.BUTTON_PRESS and
                event.button == Gdk.BUTTON_SECONDARY and pthinfo):
            path, col, cellx, celly = pthinfo
            tree_view.grab_focus()
            tree_view.set_cursor(path, col, 0)
            self._view_context_menu(event)
            return True

        return False

    def _view_context_menu(self, event):
        show = self._engine.get_show_info(self._selected_show)

        menu = Gtk.Menu()
        mb_play = Gtk.ImageMenuItem('Play Next',
                                    Gtk.Image.new_from_icon_name(
                                        "media-playback-start", Gtk.IconSize.MENU))
        mb_play.connect("activate",
                        self._on_mb_activate,
                        ShowEventType.PLAY_NEXT)
        mb_info = Gtk.MenuItem("Show details...")
        mb_info.connect("activate",
                        self._on_mb_activate,
                        ShowEventType.DETAILS)
        mb_web = Gtk.MenuItem("Open web site")
        mb_web.connect("activate",
                       self._on_mb_activate,
                       ShowEventType.OPEN_WEBSITE)
        mb_folder = Gtk.MenuItem("Open containing folder")
        mb_folder.connect("activate",
                          self._on_mb_activate,
                          ShowEventType.OPEN_FOLDER)
        mb_copy = Gtk.MenuItem("Copy title to clipboard")
        mb_copy.connect("activate",
                        self._on_mb_activate,
                        ShowEventType.COPY_TITLE)
        mb_alt_title = Gtk.MenuItem("Set alternate title...")
        mb_alt_title.connect("activate",
                             self._on_mb_activate,
                             ShowEventType.CHANGE_ALTERNATIVE_TITLE)
        mb_delete = Gtk.ImageMenuItem('Delete',
                                      Gtk.Image.new_from_icon_name(
                                          "edit-delete", Gtk.IconSize.MENU))
        mb_delete.connect("activate",
                          self._on_mb_activate,
                          ShowEventType.REMOVE)

        menu.append(mb_play)

        menu_eps = self._build_episode_menu(show)

        mb_playep = Gtk.MenuItem("Play episode")
        mb_playep.set_submenu(menu_eps)
        mb_playep.set_sensitive(bool(menu_eps.get_children()))
        menu.append(mb_playep)

        menu.append(mb_info)
        menu.append(mb_web)
        menu.append(mb_folder)
        menu.append(Gtk.SeparatorMenuItem())
        menu.append(mb_copy)
        menu.append(mb_alt_title)
        menu.append(Gtk.SeparatorMenuItem())
        menu.append(mb_delete)

        menu.show_all()
        menu.popup_at_pointer(event)

    def _build_episode_menu(self, show):
        library_episodes = set(self._engine.library().get(show['id'], ()))
        total = show['total'] or max(
            show['my_progress'],
            *library_episodes,
            utils.estimate_aired_episodes(show)
        )
        next_ep = show['my_progress'] + 1

        menu_eps = Gtk.Menu()
        for i in range(1, total + 1):
            mb_playep = Gtk.CheckMenuItem(str(i))
            if i == next_ep:
                mb_playep.set_label(str(i) + " - Next")
                menu_eps.set_focus_child(mb_playep)
            if i >= next_ep:
                mb_playep.set_margin_left(10)
            mb_playep.set_active(i in library_episodes)
            mb_playep.set_draw_as_radio(True)
            mb_playep.connect("activate",
                              self._on_mb_activate,
                              ShowEventType.PLAY_EPISODE, i)
            menu_eps.append(mb_playep)

        return menu_eps

    def _on_mb_activate(self, menu_item, event_type, data=None):
        data = (self._selected_show,) if data is None else (
            self._selected_show, data)
        self.emit('show-action', event_type, data)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt6 import QtGui

from trackma import utils


class FilterBar:
    """
    Constants relating to filter bar settings can live here.
    """
    # Position
    PositionHidden = 0
    PositionAboveLists = 1
    PositionBelowLists = 2


def getIcon(icon_name):
    fallback = QtGui.QIcon(utils.DATADIR + '/qtui/{}.png'.format(icon_name))
    return QtGui.QIcon.fromTheme(icon_name, fallback)


def getColor(colorString):
    # Takes a color string in either #RRGGBB format or group,role format (using QPalette int values)
    if colorString[0] == "#":
        return QtGui.QColor(colorString)
    else:
        (group, role) = [int(i) for i in colorString.split(',')]
        if (0 <= group <= 2) and (0 <= role <= 19):
            return QtGui.QColor(QtGui.QPalette().color(QtGui.QPalette.ColorGroup(group), QtGui.QPalette.ColorRole(role)))
        else:
            # Failsafe - return black
            return QtGui.QColor()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re
from decimal import Decimal

NO_SUBBER = '###NO#SUBBER#HERE###'

BRACKET_PAIRS = [
    # to be inserted in regular expressions
    (r'\[', r'\]'),
    (r'\(', r'\)'),
    (r'\{', r'\}'),
]


class AnimeInfoExtractor:
    """
    Extracts lots of information about anime from filename alone
    @author: Tyris
    @author: Uriziel
    @note: Thanks for Tyris for providing easy way to extract data
    @note: Does more than what's needed, but that behaviour is well tested
    """

    def __init__(self, msg, filename):
        self.msg = msg.with_classname('Parser')
        self.originalFilename = filename
        self.resolution = ''
        self.hash = ''
        self.subberTag = ''
        self.videoType = []
        self.audioType = []
        self.releaseSource = []
        self.extension = ''
        self.episodeStart = None
        self.episodeEnd = None
        self.volumeStart = None
        self.volumeEnd = None
        self.version = 1
        self.name = ''
        self.pv = -1
        self.season = None
        self._processFilename()

    def getName(self):
        return self.name

    def getEpisodeNumbers(self, force_numbers=False):
        ep_start = self.episodeStart
        ep_end = self.episodeEnd
        if force_numbers:
            if ep_start is None:
                ep_start = 1
            if ep_end is None:
                ep_end = ep_start
            ep_start = int(ep_start)
            ep_end = int(ep_end)
        return ep_start, ep_end

    def getEpisode(self):
        ep = self.episodeStart if self.episodeEnd is None else self.episodeEnd
        ep = ep if ep is not None else 1

        return int(ep)

    def __extractExtension(self, filename):
        m = re.search(r"\.(\w{3})$", filename)
        if m:
            self.extension = m.group(1)
            filename = filename[:-4]
        return filename

    def __cleanUpSpaces(self, filename):
        characters = "_.-"
        for char in characters:
            if ' ' in filename:
                break
            filename = re.sub(r'([^{0}]){0}(?=[^{0}]|$)'.format(re.escape(char)), r'\1 ', filename)
        return filename

    def __extractSpecialTags(self, filename):
        tags = {'video': r'H\.?264|x264|AVC|XviD|DivX|H\.?265|HEVC|AV1',
                'audio': r'AC3|AAC|MP3|FLAC|E-?AC-?3|Opus|DTS(?:-HD)?|TrueHD|L?PCM',
                'source': r'TV|DVD|Blu-?Ray|BD|BDMV|www|WEB(?:-DL)?'}
        for k, tag_re in tags.items():
            m = re.search(r'[\(\[][^\)\]]*?\b(' + tag_re + r')\b',
                          filename,
                          flags=re.IGNORECASE)
            if m:
                if k == 'video':
                    self.videoType.append(m.group(1))
                elif k == 'audio':
                    self.audioType.append(m.group(1))
                elif k == 'source':
                    self.releaseSource.append(m.group(1))
                # remove the match
                filename = filename[:m.start(1)] + NO_SUBBER + filename[m.end(1):]
        return filename

    def __extractVideoProfile(self, filename):
        # Check for 8bit/10bit/Hi444PP
        h264_profiles = [
            (['H264', 'Hi10P'], ['Hi10P', 'Hi10', '10bit', '10 bit', '10-bit', 'YUV420P10']),
            (['H264', '8bit'], ['8bit', '8-bit']),
            (['H264', 'Hi444PP'], ['Hi444PP', 'YUV444P10']),
        ]
        for to_add, tags in h264_profiles:
            for tag in tags:
                if tag in filename:
                    self.videoType = to_add
                    # Don't replace Hi10 because it's a subber name
                    if tag != 'Hi10':
                        filename = filename.replace(tag, '')
                    return filename
        return filename

    def __extractResolution(self, filename):
        # Match 3 or 4 chars followed by p, i, or x and 3 or 4 more chars, surrounded by any non-alphanumeric chars
        m = re.search(r'\b(\d{3,4}(?:p|i|x\d{3,4}))\b', filename)
        if m:
            self.resolution = m.group(1)
            return filename[:m.start(1)] + filename[m.end(1):]
        # HD/SD in brackets or after an episode number
        m = re.search(r'(?:[\[\(]|\d{1,3})\s*([HS]D)(TV)?\b', filename)
        if m:
            self.resolution = m.group(1)
            if m.group(2):
                self.releaseSource.append(m.group(2))
            return filename[:m.start(1)] + filename[m.end():]
        # HD/SD at the end
        m = re.search(r'\b([HS]D)(TV)?$', filename)
        if m:
            self.resolution = m.group(1)
            if m.group(2):
                self.releaseSource.append(m.group(2))
            return filename[:m.start()] + filename[m.end():]

        return filename

    def __extractHash(self, filename):
        # Match anything in square or round brackets that is 8 hex digits
        m = re.search(r'[\(\[]([A-Fa-f0-9]{8})[\)\]]', filename)
        if m:
            self.hash = m.group(1)
            filename = filename[:m.start()] + filename[m.end():]
        return filename

    def __checkIfRemux(self, filename):
        m = re.search(r'[\(\[][^\)\]]*?(Remux)\b', filename, flags=re.IGNORECASE)
        return True if m else False

    def __cleanUpBrackets(self, filename):
        # Can get rid of the brackets that won't contain subber
        filename = re.sub(r'\([^\)]*?' + NO_SUBBER + r'.*?\)', '', filename)
        filename = re.sub(r'\[[^\]]*?' + NO_SUBBER + r'.*?\]', '', filename)
        # Strip any empty sets of brackets, unless they are at the beginning
        filename = re.sub(r'(?!^)\[\W*?\]|\(\W*?\)', '', filename)
        return filename

    def __extractSubber(self, filename, remux):
        # Extract the subber from square brackets (or round failing that)
        for opening, closing in BRACKET_PAIRS:
            m = re.search(r'{0}([^\. ].*?){1}'.format(opening, closing), filename)
            if m:
                self.subberTag = m.group(1)
                filename = filename[:m.start()] + filename[m.end():]
                break
        # Add the remux string if this was a remux and it's not found in the subber tag
        if remux and 'remux' not in self.subberTag.lower():
            # refind remux and remove it
            m = re.search(
                r'[\(\[][^\)\]]*?(Remux)\b', filename, flags=re.IGNORECASE)
            if m:
                filename = filename[:m.start(1)] + filename[m.end(1):]
            if self.subberTag:
                self.subberTag = self.subberTag + '-Remux'
            else:
                self.subberTag = 'Remux'
        return filename

    def __extractVersion(self, filename):
        # Extract the version number (limit at v7 since V8 is possible in a title...)
        m = re.search(r'(?:[^a-zA-Z])v([0-7])\b',
                      filename, flags=re.IGNORECASE)
        if m:
            self.version = int(m.group(1))
            filename = filename[:m.start(1) - 1] + filename[m.end(1):]
        return filename

    def __extractVolumeIfPack(self, filename, title_len):
        # Check if this is a volume pack - only relevant for no extension
        if not self.extension:
            m = re.search(
                r'\b(?:vol(?:ume)?\.? ?)(\d{1,3})(?: ?- ?(?:vol(?:ume)?\.? ?)?(\d{1,3}))?\b',
                filename,
                flags=re.IGNORECASE)
            if m:
                self.volumeStart = int(m.group(1))
                if m.group(2):
                    self.volumeEnd = int(m.group(2))
                filename = filename[:m.start()] + filename[m.end():]
                title_len = m.start()
        return filename, title_len

    def __extractPv(self, filename):
        # Check if this is a PV release (not relevant if it's a pack)
        m = re.search(r' PV ?(\d)?(?:[^a-zA-Z0-9]|$)', filename)
        if not self.volumeStart and m:
            self.pv = 0
            if m.group(1):
                self.pv = int(m.group(1))
            filename = filename[:m.start(0)]
        return filename

    def __extractEpisodeNumbers(self, filename):
        # First check for concurrent episodes (with a + or &)
        m = re.search(
            r'\b(?:S(?:\.|eason)?(\d+)\s*)?'
            r'(?:E\.?|Ep(?:i|isode)?s?[ .]?)?(\d{1,4})[\+\&](\d{1,4})\b',
            filename,
            flags=re.IGNORECASE,
        )
        if m:
            start = int(m.group(2))
            end = int(m.group(3))
            if end == start + 1:
                if m.group(1):
                    self.season = int(m.group(1))
                self.episodeStart = start
                self.episodeEnd = end
                return filename[:m.start()]

        # Check for multiple episodes (with a -)
        ep_search_string = (
            r'\b(?:S(?:\.|eason)?(\d+)\s*)?'
            r'(?:E\.?|Ep(?:i|isode)?[ .]?)?'
            r'((?:\d{1,3}|1[0-8]\d{2})(?:\.\d{1})?)'
            # Only allow spaces around the hyphen when we are likely to have a pack
            + (r'-' if self.extension else r' ?- ?')
            + r'(\d{1,4}(?:\.\d{1})?)\b'
        )
        m = re.search(ep_search_string, filename, flags=re.IGNORECASE)
        if m:
            if m.group(1):
                self.season = int(m.group(1))
            self.episodeStart = Decimal(m.group(2))
            self.episodeEnd = Decimal(m.group(3))
            return filename[:m.start()]

        # Check if there is an episode specifier
        m = re.search(
            r'\b(?:S(?:\.|eason)?(\d+)\s*)?(?:E\.?|Ep(?:i|isode)?[ .]?)(\d+(?:\.\d)?)(?:\b|v)',
            filename,
            flags=re.IGNORECASE
        )
        if m:
            if m.group(1):
                self.season = int(m.group(1))
            self.episodeStart = Decimal(m.group(2))
            return filename[:m.start()]

        # Check any remaining lonely numbers as episode (towards the end has priority)
        # First try outside brackets
        m = re.search(r'.*[^\.\[\(]\b((?:\d{1,3}|1[0-8]\d{2})(?:\.\d)?)(?:[\[({]|\s*$|\s+\W)',
                      filename)
        if m:
            self.episodeStart = Decimal(m.group(1))
            return filename[:m.start(1)]

        # then allow brackets
        m = re.search(r'.*[\[\(]((?:\d{1,3}|1[0-8]\d{2})(?:\.\d)?)(?:[\])}]|\s*$|\s+\W)', filename)
        if m:
            self.episodeStart = Decimal(m.group(1))
            return filename[:m.start(1)]
        return filename

    def __extractShowName(self, filename):
        # Unfortunately it's very hard to know if there should be brackets in the title.
        # We really should strip brackets, though, so sorry to anything with brackets in the title.
        # We don't strip years or the whole title, however.
        for opening, closing in BRACKET_PAIRS:
            m = re.search(r'{0}((?!\d{{4}}{1}).*?){1}'.format(opening, closing), filename)
            if m and m.end() - m.start() < len(filename):
                filename = filename[:m.start()] + filename[m.end():]
        filename = re.sub(r'  .*', '', filename)
        # Strip any unclosed brackets and anything after them
        for opening, closing in BRACKET_PAIRS:
            filename = re.sub(r'{}r[^{}].*$'.format(opening, closing), '', filename)
        self.name = re.sub(r'( - *)+$', '', filename.strip(' '))
        # If we have a subber but no title!? then it must have been a title...
        if self.name == '' and self.subberTag != '':
            self.name = self.subberTag
            self.subberTag = ''
        # Reappend season number
        if self.season and self.season > 1:
            self.name += " Season {}".format(self.season)

    def _processFilename(self):
        filename = self.originalFilename
        filename = self.__extractExtension(filename)
        filename = self.__cleanUpSpaces(filename)
        filename = self.__extractSpecialTags(filename)
        filename = self.__extractVideoProfile(filename)
        filename = self.__extractResolution(filename)
        filename = self.__extractHash(filename)
        remux = self.__checkIfRemux(filename)
        filename = self.__cleanUpBrackets(filename)
        filename = self.__extractSubber(filename, remux)
        filename = self.__extractVersion(filename)
        # Store the possible length of the title
        title_len = len(filename)
        filename, title_len = self.__extractVolumeIfPack(filename, title_len)
        filename = self.__extractPv(filename)
        if self.pv == -1:
            filename = self.__extractEpisodeNumbers(filename)
        # Truncate remainder to title length if needed (for where volume was found)
        filename = filename[:title_len]
        # Strip any trailing opening brackets
        filename = filename.rstrip('([{')
        self.__extractShowName(filename)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from enum import Enum
import os
import re
import sys
import threading
import urllib.parse
import asyncio
from typing import Dict, Union
from dataclasses import dataclass

from jeepney import HeaderFields, Properties, DBusAddress
from jeepney.io.asyncio import open_dbus_router, DBusRouter, Proxy
from jeepney.bus_messages import message_bus, MatchRule

from trackma import utils
from trackma.tracker import tracker


BUS_NAMESPACE = 'org.mpris.MediaPlayer2'
PATH_MPRIS = '/org/mpris/MediaPlayer2'
IFACE_PROPERTIES = 'org.freedesktop.DBus.Properties'
IFACE_MPRIS = 'org.mpris.MediaPlayer2'
IFACE_MPRIS_PLAYER = IFACE_MPRIS + '.Player'


class PlaybackStatus(str, Enum):
    PLAYING = 'Playing'
    PAUSED = 'Paused'
    STOPPED = 'Stopped'


async def name_owner_watcher(router, tracker):
    # Select name change signals for the well-known mpris service name.
    # We only consider players with such a well-known name
    # and thus treat any changes to the owned name
    # as a creation and a disappearance of a service.
    # Signals will be published by the service's unique name,
    # which is why we need to track it.
    match_rule = MatchRule(
        type='signal',
        sender=message_bus.bus_name,
        interface=message_bus.interface,
        member='NameOwnerChanged',
        path=message_bus.object_path,
    )
    match_rule.add_arg_condition(0, BUS_NAMESPACE, kind='namespace')

    message_proxy = Proxy(message_bus, router)
    await message_proxy.AddMatch(match_rule)

    with router.filter(match_rule, bufsize=20) as queue:
        while True:
            # https://dbus.freedesktop.org/doc/dbus-specification.html#bus-messages-name-owner-changed
            msg = await queue.get()
            (wellknown_name, old_name, new_name) = msg.body
            if new_name:
                await tracker.on_bus_added(router, wellknown_name, new_name)
            elif old_name:
                tracker.on_bus_removed(wellknown_name, new_name)


def safe_get_dbus_value(pair, type_):
    """Assert the value's type before returning it None-safely."""
    if pair is None:
        return None
    if pair[0] != type_:
        raise TypeError(f"Expected type {type_!r} but found {pair[0]!r}")
    return pair[1]


async def properties_watcher(router, tracker):
    # Select PropertiesChanged signals for the mpris-player subinterface
    # for all senders.
    match_rule = MatchRule(
        type='signal',
        sender=None,
        interface=IFACE_PROPERTIES,
        member='PropertiesChanged',
        path=PATH_MPRIS,
    )
    match_rule.add_arg_condition(0, IFACE_MPRIS_PLAYER)

    message_proxy = Proxy(message_bus, router)
    await message_proxy.AddMatch(match_rule)

    with router.filter(match_rule, bufsize=10) as queue:
        while True:
            msg = await queue.get()
            handle_properties_changed(tracker, msg)


def handle_properties_changed(tracker, msg):
    # https://dbus.freedesktop.org/doc/dbus-specification.html#standard-interfaces-properties
    (_, changed_properties, _) = msg.body
    sender = msg.header.fields[HeaderFields.sender]
    try:
        playback_status = safe_get_dbus_value(changed_properties.get('PlaybackStatus'), 's')
        if playback_status:
            tracker.on_playback_status_change(sender, playback_status)

        metadata = safe_get_dbus_value(changed_properties.get('Metadata'), 'a{sv}')
        if metadata and ('xesam:title' in metadata or 'xesam:url' in metadata):
            title = safe_get_dbus_value(metadata.get('xesam:title'), 's')
            url = safe_get_dbus_value(metadata.get('xesam:url'), 's')
            tracker.on_filename_change(sender, title, url)
    except TypeError as e:
        tracker.msg.warn(f"Failed to read properties for {sender}: {e}")


async def collect_names(router) -> Dict[str, str]:
    """Find all active buses in the known namespace mapped by their unique name."""
    message_proxy = Proxy(message_bus, router)
    resp = await message_proxy.ListNames()
    names = [name for name in resp[0] if name.startswith(BUS_NAMESPACE)]
    unique_names = await asyncio.gather(*(message_proxy.GetNameOwner(name) for name in names))
    name_map = {un[0]: n for un, n in zip(unique_names, names)}
    return name_map


@dataclass
class Player:
    router: DBusRouter
    wellknown_name: str
    unique_name: str
    playback_status: Union[str, None] = None
    title: Union[str, None] = None
    url: Union[str, None] = None

    @classmethod
    async def new(cls, router, wellknown_name, unique_name):
        player = Player(router, wellknown_name, unique_name)
        await asyncio.gather(
            player.update_filename(),
            player.update_playback_status(),
        )
        return player

    @property
    def filename(self):
        if self.title and len(self.title) > 5:
            return self.title
        elif self.url:
            # TODO : Support for full path
            return os.path.basename(urllib.parse.unquote_plus(self.url))
        return self.title

    async def update_filename(self):
        msg = self._player_properties.get('Metadata')
        reply = await self.router.send_and_get_reply(msg)
        metadata = safe_get_dbus_value(reply.body[0], 'a{sv}')
        if metadata:
            self.title = safe_get_dbus_value(metadata.get('xesam:title'), 's')
            self.url = safe_get_dbus_value(metadata.get('xesam:url'), 's')
        else:
            self.title = None
            self.url = None

    async def update_playback_status(self):
        msg = self._player_properties.get('PlaybackStatus')
        reply = await self.router.send_and_get_reply(msg)
        self.playback_status = safe_get_dbus_value(reply.body[0], 's')

    async def get_position(self):
        msg = self._player_properties.get('Position')
        reply = await self.router.send_and_get_reply(msg)
        return safe_get_dbus_value(reply.body[0], 'x')

    @property
    def _player_properties(self):
        address = DBusAddress(PATH_MPRIS, bus_name=self.unique_name, interface=IFACE_MPRIS_PLAYER)
        return Properties(address)


class MprisTracker(tracker.TrackerBase):

    name = 'Tracker (MPRIS)'

    def __init__(self, *args, **kwargs):
        # The `TrackerBase.__init__` spawns a new thread
        # for `observe`.
        self.initialized = threading.Event()
        super().__init__(*args, **kwargs)

        self.re_players = re.compile(self.config['tracker_process'])
        # Map "OwnerName"s to Player instances
        # for monitoring when one appears or disappears
        # via the `NameOwnerChanged` signal.
        self.players = {}
        self.timing = False
        self.active_player = None
        self.initialized.set()

    def update_list(self, *args, **kwargs):
        super().update_list(*args, **kwargs)
        if self.last_state != utils.Tracker.PLAYING:
            # Re-check if we have any player with a valid show running after a list update
            self.last_filename = None
            self.find_playing_player()

    def observe(self, config, watch_dirs):
        self.msg.info("Using MPRIS.")
        self.initialized.wait()
        asyncio.run(self.observe_async())

    async def observe_async(self):
        async with open_dbus_router() as router:
            name_owner_watcher_task = asyncio.create_task(name_owner_watcher(router, self))
            properties_watcher_task = asyncio.create_task(properties_watcher(router, self))
            timer_task = asyncio.create_task(self._timer())
            tasks = [name_owner_watcher_task, properties_watcher_task, timer_task]

            name_map = await collect_names(router)
            self.players = {
                unique_name: await Player.new(router, wellknown_name, unique_name)
                for unique_name, wellknown_name in name_map.items()
                if self.valid_player(wellknown_name)
            }
            ignored_players = [name for name in name_map.values() if not self.valid_player(name)]
            self.msg.debug(f"Ignoring players: {ignored_players}")
            for player in self.players.values():
                self.msg.debug(f"Player connected: {player.wellknown_name}")
            self.find_playing_player()

            try:
                await asyncio.gather(*tasks)
            except Exception:
                self.msg.exception("Error in dbus watchers; cleaning up", sys.exc_info())
                for task in tasks:
                    task.cancel()
                await asyncio.gather(*tasks)
                # let the thread die

    def valid_player(self, wellknown_name):
        return self.re_players.search(wellknown_name)

    def find_playing_player(self) -> bool:
        # Go through all connected players in random order
        # (or not since dicts are ordered now).
        return any(
            self._update_active_player(player, probing=True)
            for player in self.players.values()
            if player.playback_status == PlaybackStatus.PLAYING
        )

    async def on_bus_added(self, router, wellknown_name, unique_name):
        if not self.valid_player(wellknown_name):
            self.msg.debug("Ignoring new bus that does not match configured player:"
                           f" {wellknown_name}")
            return
        player = await Player.new(router, wellknown_name, unique_name)
        self.msg.debug(f"Player connected: {player.wellknown_name}")
        self.players[unique_name] = player
        self._handle_player_update(player)

    def on_bus_removed(self, wellknown_name, unique_name):
        player = self.players.pop(unique_name, None)
        if not player:
            return
        self.msg.debug(f"Player disconnected: {player.wellknown_name}")
        if player == self.active_player:
            self._handle_player_stopped()
            self.active_player = None

    def on_playback_status_change(self, sender, playback_status):
        player = self.players.get(sender)
        if not player:
            return
        player.playback_status = playback_status
        self._handle_player_update(player)

    def on_filename_change(self, sender, title, url):
        player = self.players.get(sender)
        if not player:
            return
        player.title = title
        player.url = url
        self._handle_player_update(player)

    def _handle_player_update(self, player):
        if player == self.active_player:
            if player.playback_status == PlaybackStatus.STOPPED:
                self._handle_player_stopped()
            else:
                self._update_active_player(player)

        elif self.active_player and self.active_player.playback_status == PlaybackStatus.PLAYING:
            self.msg.debug("Still playing on active player; ignoring update")
            return

        elif player.playback_status == PlaybackStatus.PLAYING:
            self._update_active_player(player)

    def _update_active_player(self, player: Player, probing=False) -> bool:
        is_new_player = player != self.active_player

        (state, show_tuple) = (None, None)
        previous_last_filename = self.last_filename
        new_show = False
        if player.filename != self.last_filename:
            (state, show_tuple) = self._get_playing_show(player.filename)
            if state in [utils.Tracker.UNRECOGNIZED, utils.Tracker.NOT_FOUND]:
                self.msg.debug("Video not recognized")
            elif state == utils.Tracker.NOVIDEO:
                self.msg.debug("No video loaded")
            else:
                new_show = True

        if is_new_player and not new_show:
            if probing:
                # Ignore this 'new' player & restore `last_filename`
                # since we're just looking for a new player candidate.
                # (This is a hack but a proper fix needs larger refactoring
                # involving the parent class.)
                self.last_filename = previous_last_filename
            return False

        if is_new_player and new_show:
            self.msg.debug(f"Setting active player: {player.wellknown_name}")
            self.active_player = player

        if state:
            self.msg.debug(f"New tracker status: {state} (previously: {self.last_state})")
            self.update_show_if_needed(state, show_tuple)

        if player.playback_status == PlaybackStatus.PLAYING:
            self._start_timer()
        else:
            self._stop_timer()

        return True

    def _handle_player_stopped(self):
        # Active player got closed!
        if self.active_player:
            self.msg.debug(f"Clearing active player: {self.active_player.wellknown_name}")
        self.active_player = None
        self.view_offset = None

        if not self.find_playing_player():
            (state, show_tuple) = self._get_playing_show(None)
            self.update_show_if_needed(state, show_tuple)
            self._stop_timer()

    def _start_timer(self):
        self.resume_timer()
        if not self.timing:
            self.timing = True
            self.msg.debug("MPRIS timer started.")

    def _stop_timer(self):
        self.pause_timer()
        if self.timing:
            self.timing = False
            self.msg.debug("MPRIS timer paused.")

    async def _timer(self):
        while True:
            if self.timing:
                await self._on_tick()
            await asyncio.sleep(1, True)

    async def _on_tick(self):
        if self.active_player:
            self.view_offset = int(await self.active_player.get_position()) / 1000
        if self.last_show_tuple:
            self.update_timer(self.last_state, self.last_show_tuple)
        if self.last_updated:
            self._stop_timer()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import base64
import os

from PyQt6 import QtCore, QtGui
from PyQt6.QtGui import QAction, QActionGroup
from PyQt6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QComboBox,
                             QDoubleSpinBox, QFormLayout, QHBoxLayout, QHeaderView, QInputDialog, QLabel, QLineEdit,
                             QMainWindow, QMenu, QMessageBox, QProgressBar, QPushButton, QSpinBox, QStyle,
                             QStyleOptionButton, QSystemTrayIcon, QTabBar, QToolButton, QVBoxLayout, QWidget)

from trackma import messenger
from trackma import utils
from trackma.accounts import AccountManager
from trackma.ui.qt.accounts import AccountDialog
from trackma.ui.qt.add import AddDialog
from trackma.ui.qt.details import DetailsDialog
from trackma.ui.qt.settings import SettingsDialog
from trackma.ui.qt.util import FilterBar, getIcon
from trackma.ui.qt.widgets import ShowsTableView
from trackma.ui.qt.workers import EngineWorker, ImageWorker

class MainWindow(QMainWindow):
    """
    Main GUI class

    """
    debug = False
    config = None
    tray = None
    mediainfo = None
    accountman = None
    accountman_widget = None
    worker = None
    image_worker = None
    started = False
    selected_show_id = None
    show_lists = None
    finish = False
    was_maximized = False

    def __init__(self, debug=False):
        QMainWindow.__init__(self, None)
        self.debug = debug

        # Load QT specific configuration
        self.configfile = utils.to_config_path('ui-qt.json')
        self.config = utils.parse_config(self.configfile, utils.qt_defaults)

        # Build UI
        if os.name != "nt":
            QApplication.setWindowIcon(QtGui.QIcon(utils.DATADIR + '/icon.png'))
        else:
            QApplication.setWindowIcon(QtGui.QIcon(utils.DATADIR + '/icon.ico'))
        self.setWindowTitle('Trackma-qt')

        self.accountman = AccountManager()

        # Go directly into the application if a default account is set
        # Open the selection dialog otherwise
        default = self.accountman.get_default()
        if default:
            self.start(default)
        else:
            self.accountman_create()
            self.accountman_widget.show()

    def accountman_create(self):
        self.accountman_widget = AccountDialog(None, self.accountman)
        self.accountman_widget.selected.connect(self.accountman_selected)

    def accountman_selected(self, account_num, remember):
        account = self.accountman.get_account(account_num)

        if remember:
            self.accountman.set_default(account_num)
        else:
            self.accountman.set_default(None)

        if self.started:
            self.reload(account)
        else:
            self.start(account)

    def start(self, account):
        """
        Start engine and everything

        """
        # Workers
        self.worker = EngineWorker()
        self.account = account

        # Get API specific configuration
        self.api_config = self._get_api_config(account['api'])

        # Timers
        self.image_timer = QtCore.QTimer()
        self.image_timer.setInterval(500)
        self.image_timer.setSingleShot(True)
        self.image_timer.timeout.connect(self.s_download_image)

        self.busy_timer = QtCore.QTimer()
        self.busy_timer.setInterval(100)
        self.busy_timer.setSingleShot(True)
        self.busy_timer.timeout.connect(self.s_busy)

        # Build menus
        self.action_play_next = QAction(
            getIcon('media-playback-start'), 'Play &Next', self)
        self.action_play_next.setStatusTip('Play the next unwatched episode.')
        self.action_play_next.setShortcut('Ctrl+N')
        self.action_play_next.triggered.connect(lambda: self.s_play(True))
        self.action_play_dialog = QAction('Play Episode...', self)
        self.action_play_dialog.setStatusTip('Select an episode to play.')
        self.action_play_dialog.triggered.connect(self.s_play_number)
        self.action_details = QAction('Show &details...', self)
        self.action_details.setStatusTip(
            'Show detailed information about the selected show.')
        self.action_details.triggered.connect(self.s_show_details)
        self.action_altname = QAction('Change &alternate name...', self)
        self.action_altname.setStatusTip(
            'Set an alternate title for the tracker.')
        self.action_altname.triggered.connect(self.s_altname)
        action_play_random = QAction('Play &random show', self)
        action_play_random.setStatusTip(
            'Pick a random show with a new episode and play it.')
        action_play_random.setShortcut('Ctrl+R')
        action_play_random.triggered.connect(self.s_play_random)
        self.action_add = QAction(
            getIcon('edit-find'), 'Search/Add from Remote', self)
        self.action_add.setShortcut('Ctrl+A')
        self.action_add.triggered.connect(self.s_add)
        self.action_delete = QAction(getIcon('edit-delete'), '&Delete', self)
        self.action_delete.setStatusTip('Remove this show from your list.')
        self.action_delete.setShortcut(QtCore.Qt.Key.Key_Delete)
        self.action_delete.triggered.connect(self.s_delete)
        action_quit = QAction(getIcon('application-exit'), '&Quit', self)
        action_quit.setShortcut('Ctrl+Q')
        action_quit.setStatusTip('Exit Trackma.')
        action_quit.triggered.connect(self._exit)

        self.action_sync = QAction('&Sync', self)
        self.action_sync.setStatusTip(
            'Send changes and then retrieve remote list')
        self.action_sync.setShortcut('Ctrl+S')
        self.action_sync.triggered.connect(lambda: self.s_send(True))
        self.action_send = QAction('S&end changes', self)
        self.action_send.setShortcut('Ctrl+E')
        self.action_send.setStatusTip(
            'Upload any changes made to the list immediately.')
        self.action_send.triggered.connect(self.s_send)
        self.action_retrieve = QAction('Re&download list', self)
        self.action_retrieve.setShortcut('Ctrl+D')
        self.action_retrieve.setStatusTip(
            'Discard any changes made to the list and re-download it.')
        self.action_retrieve.triggered.connect(self.s_retrieve)
        action_scan_library = QAction('Rescan &Library (quick)', self)
        action_scan_library.setShortcut('Ctrl+L')
        action_scan_library.triggered.connect(self.s_scan_library)
        action_rescan_library = QAction('Rescan &Library (full)', self)
        action_rescan_library.triggered.connect(self.s_rescan_library)
        action_open_folder = QAction('Open containing folder', self)
        action_open_folder.triggered.connect(self.s_open_folder)

        self.action_reload = QAction('Switch &Account', self)
        self.action_reload.setStatusTip('Switch to a different account.')
        self.action_reload.triggered.connect(self.s_switch_account)
        action_settings = QAction('&Settings...', self)
        action_settings.triggered.connect(self.s_settings)

        action_about = QAction(getIcon('help-about'), 'About...', self)
        action_about.triggered.connect(self.s_about)
        action_about_qt = QAction('About Qt...', self)
        action_about_qt.triggered.connect(self.s_about_qt)

        menubar = self.menuBar()
        self.menu_show = menubar.addMenu('&Show')
        self.menu_show.addAction(self.action_play_next)
        self.menu_show.addAction(self.action_play_dialog)
        self.menu_show.addAction(self.action_details)
        self.menu_show.addAction(self.action_altname)
        self.menu_show.addSeparator()
        self.menu_show.addAction(action_play_random)
        self.menu_show.addSeparator()
        self.menu_show.addAction(self.action_add)
        self.menu_show.addAction(self.action_delete)
        self.menu_show.addSeparator()
        self.menu_show.addAction(action_quit)

        self.menu_play = QMenu('Play')

        # Context menu for right click on list item
        self.menu_show_context = QMenu()
        self.menu_show_context.addMenu(self.menu_play)
        self.menu_show_context.addAction(self.action_details)
        self.menu_show_context.addAction(action_open_folder)
        self.menu_show_context.addAction(self.action_altname)
        self.menu_show_context.addSeparator()
        self.menu_show_context.addAction(self.action_delete)

        # Make icons for viewed episodes
        rect = QtCore.QSize(16, 16)
        buffer = QtGui.QPixmap(rect)
        ep_icon_states = {'all': QStyle.StateFlag.State_On,
                          'part': QStyle.StateFlag.State_NoChange,
                          'none': QStyle.StateFlag.State_Off}
        self.ep_icons = {}
        for key, state in ep_icon_states.items():
            buffer.fill(QtCore.Qt.GlobalColor.transparent)
            painter = QtGui.QPainter(buffer)
            opt = QStyleOptionButton()
            opt.state = state
            self.style().drawPrimitive(QStyle.PrimitiveElement.PE_IndicatorMenuCheckMark, opt, painter)
            self.ep_icons[key] = QtGui.QIcon(buffer)
            painter.end()

        menu_list = menubar.addMenu('&List')
        menu_list.addAction(self.action_sync)
        menu_list.addSeparator()
        menu_list.addAction(self.action_send)
        menu_list.addAction(self.action_retrieve)
        menu_list.addSeparator()
        menu_list.addAction(action_scan_library)
        menu_list.addAction(action_rescan_library)
        self.menu_mediatype = menubar.addMenu('&Mediatype')
        self.mediatype_actiongroup = QActionGroup(self)
        self.mediatype_actiongroup.setExclusive(True)
        self.mediatype_actiongroup.triggered.connect(self.s_mediatype)
        menu_options = menubar.addMenu('&Options')
        menu_options.addAction(self.action_reload)
        menu_options.addSeparator()
        menu_options.addAction(action_settings)
        menu_help = menubar.addMenu('&Help')
        menu_help.addAction(action_about)
        menu_help.addAction(action_about_qt)

        # Build layout
        main_layout = QVBoxLayout()
        top_hbox = QHBoxLayout()
        main_hbox = QHBoxLayout()
        self.list_box = QVBoxLayout()
        filter_bar_box_layout = QHBoxLayout()
        self.filter_bar_box = QWidget()
        left_box = QFormLayout()
        small_btns_hbox = QHBoxLayout()

        self.show_title = QLabel('Trackma-qt')
        show_title_font = QtGui.QFont()
        show_title_font.setBold(True)
        show_title_font.setPointSize(12)
        self.show_title.setFont(show_title_font)

        self.api_icon = QLabel('icon')
        self.api_user = QLabel('user')

        top_hbox.addWidget(self.show_title, 1)
        top_hbox.addWidget(self.api_icon)
        top_hbox.addWidget(self.api_user)

        # Create main models and view
        self.notebook = QTabBar()
        self.notebook.currentChanged.connect(self.s_tab_changed)

        self.view = ShowsTableView(palette=self.config['colors'])
        self.view.context_menu = self.menu_show_context
        self.view.horizontalHeader().customContextMenuRequested.connect(
            self.s_show_menu_columns)
        self.view.horizontalHeader().sortIndicatorChanged.connect(self.s_update_sort)
        self.view.selectionModel().currentRowChanged.connect(self.s_show_selected)
        self.view.itemDelegate().setBarStyle(
            self.config['episodebar_style'], self.config['episodebar_text'])
        self.view.middleClicked.connect(lambda: self.s_play(True))
        self.view.doubleClicked.connect(self.s_show_details)
        self._apply_view()

        self.view.model().sourceModel().progressChanged.connect(self.s_set_episode)
        self.view.model().sourceModel().scoreChanged.connect(self.s_set_score)

        # Context menu for right click on list header
        self.menu_columns = QMenu()
        self.column_keys = {'id': 0,
                            'title': 1,
                            'progress': 2,
                            'score': 3,
                            'percent': 4,
                            'next_ep': 5,
                            'date_start': 6,
                            'date_end': 7,
                            'my_start': 8,
                            'my_end': 9,
                            'tag': 10,
                            'my_last_update': 12}

        for i, column_name in enumerate(self.view.model().sourceModel().columns):
            action = QAction(column_name, self, checkable=True)
            action.setData(i)
            if column_name in self.api_config['visible_columns']:
                action.setChecked(True)

            action.triggered.connect(self.s_toggle_column)
            self.menu_columns.addAction(action)

        # Create filter list
        self.show_filter = QLineEdit()
        self.show_filter.setClearButtonEnabled(True)
        self.show_filter.textChanged.connect(self.s_filter_changed)
        filter_tooltip = (
            "General Search: All fields (columns) of each show will be matched against the search term."
            "\nAdvanced Searching: A field can be specified by using its key followed by a colon"
            " e.g. 'title:My_Show date_start:2016'."
            "\n  Any field may be specified multiple times to match terms in any order e.g. 'tag:Battle+Shounen tag:Ecchi'. "
            "\n  + and _ are replaced with spaces when searching specific fields."
            "\n  If colon is used after something that is not a column key, it will treat it as a general term."
            "\n  ALL terms not attached to a field will be combined into a single general search term"
            "\n         - 'My date_end:2016 Show' will match shows that have 'My Show' in any field and 2016 in the End Date field."
            "\n  Available field keys are: "
        )
        colkeys = ', '.join(sorted(self.column_keys.keys()))
        self.show_filter.setToolTip(filter_tooltip + colkeys + '.')
        self.show_filter_invert = QCheckBox()
        self.show_filter_invert.stateChanged.connect(self.s_filter_changed)
        self.show_filter_casesens = QCheckBox()
        self.show_filter_casesens.stateChanged.connect(self.s_filter_changed)

        if self.config['remember_geometry']:
            self.resize(self.config['last_width'], self.config['last_height'])
            self.move(self.config['last_x'], self.config['last_y'])
        else:
            self.resize(740, 480)

        spinbox_width = 75
        self.show_image = QLabel('Trackma-qt')
        self.show_image.setFixedHeight(149)
        self.show_image.setMinimumWidth(100)
        self.show_image.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        show_progress_label = QLabel('Progress:')
        self.show_progress = QSpinBox()
        self.show_progress.setMinimumWidth(spinbox_width)
        self.show_progress_bar = QProgressBar()
        self.show_progress_btn = QPushButton('Update')
        self.show_progress_btn.setToolTip(
            'Set number of episodes watched to the value entered above')
        self.show_progress_btn.clicked.connect(self.s_set_episode)
        self.show_play_btn = QToolButton()
        self.show_play_btn.setIcon(getIcon('media-playback-start'))
        self.show_play_btn.setToolTip(
            'Play the next unwatched episode\nHold to play other episodes')
        self.show_play_btn.clicked.connect(lambda: self.s_play(True))
        self.show_play_btn.setMenu(self.menu_play)
        self.show_inc_btn = QToolButton()
        self.show_inc_btn.setIcon(getIcon('list-add'))
        self.show_inc_btn.setShortcut('Ctrl+Right')
        self.show_inc_btn.setToolTip('Increment number of episodes watched')
        self.show_inc_btn.clicked.connect(self.s_plus_episode)
        self.show_dec_btn = QToolButton()
        self.show_dec_btn.setIcon(getIcon('list-remove'))
        self.show_dec_btn.clicked.connect(self.s_rem_episode)
        self.show_dec_btn.setShortcut('Ctrl+Left')
        self.show_dec_btn.setToolTip('Decrement number of episodes watched')
        show_score_label = QLabel('Score:')
        self.show_score = QDoubleSpinBox()
        self.show_score.setMinimumWidth(spinbox_width)
        self.show_score_btn = QPushButton('Set')
        self.show_score_btn.setToolTip('Set score to the value entered above')
        self.show_score_btn.clicked.connect(self.s_set_score)
        self.show_tags_btn = QPushButton('Edit Tags...')
        self.show_tags_btn.setToolTip(
            'Open a dialog to edit your tags for this show')
        self.show_tags_btn.clicked.connect(self.s_set_tags)
        self.show_status = QComboBox()
        self.show_status.setToolTip('Change your watching status of this show')
        self.show_status.currentIndexChanged.connect(self.s_set_status)

        small_btns_hbox.addWidget(self.show_dec_btn)
        small_btns_hbox.addWidget(self.show_play_btn)
        small_btns_hbox.addWidget(self.show_inc_btn)
        small_btns_hbox.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        left_box.addRow(self.show_image)
        left_box.addRow(self.show_progress_bar)
        left_box.addRow(small_btns_hbox)
        left_box.addRow(show_progress_label)
        left_box.addRow(self.show_progress, self.show_progress_btn)
        left_box.addRow(show_score_label)
        left_box.addRow(self.show_score, self.show_score_btn)
        left_box.addRow(self.show_status)
        left_box.addRow(self.show_tags_btn)

        filter_bar_box_layout.addWidget(QLabel('Filter:'))
        filter_bar_box_layout.addWidget(self.show_filter)
        filter_bar_box_layout.addWidget(QLabel('Invert'))
        filter_bar_box_layout.addWidget(self.show_filter_invert)
        filter_bar_box_layout.addWidget(QLabel('Case Sensitive'))
        filter_bar_box_layout.addWidget(self.show_filter_casesens)
        self.filter_bar_box.setLayout(filter_bar_box_layout)

        if self.config['filter_bar_position'] is FilterBar.PositionHidden:
            self.list_box.addWidget(self.notebook)
            self.list_box.addWidget(self.view)
            self.filter_bar_box.hide()
        elif self.config['filter_bar_position'] is FilterBar.PositionAboveLists:
            self.list_box.addWidget(self.filter_bar_box)
            self.list_box.addWidget(self.notebook)
            self.list_box.addWidget(self.view)
        elif self.config['filter_bar_position'] is FilterBar.PositionBelowLists:
            self.list_box.addWidget(self.notebook)
            self.list_box.addWidget(self.view)
            self.list_box.addWidget(self.filter_bar_box)

        main_hbox.addLayout(left_box)
        main_hbox.addLayout(self.list_box, 1)

        main_layout.addLayout(top_hbox)
        main_layout.addLayout(main_hbox)

        self.main_widget = QWidget(self)
        self.main_widget.setLayout(main_layout)
        self.setCentralWidget(self.main_widget)

        # Statusbar
        self.status_text = QLabel('Trackma-qt')
        self.tracker_text = QLabel('Tracker: N/A')
        self.tracker_text.setMinimumWidth(120)
        self.queue_text = QLabel('Unsynced items: N/A')
        self.statusBar().addWidget(self.status_text, 1)
        self.statusBar().addPermanentWidget(self.tracker_text)
        self.statusBar().addPermanentWidget(self.queue_text)

        # Tray icon
        tray_menu = QMenu(self)
        action_hide = QAction('Show/Hide', self)
        action_hide.triggered.connect(self.s_hide)
        tray_menu.addAction(action_hide)
        tray_menu.addAction(action_quit)

        self.tray = QSystemTrayIcon(self.windowIcon())
        self.tray.setContextMenu(tray_menu)
        self.tray.activated.connect(self.s_tray_clicked)
        self._apply_tray()

        # Connect worker signals
        self.worker.changed_status.connect(self.ws_changed_status)
        self.worker.raised_error.connect(self.error)
        self.worker.raised_fatal.connect(self.fatal)
        self.worker.changed_show.connect(self.ws_changed_show)
        self.worker.changed_show_status.connect(self.ws_changed_show_status)
        self.worker.changed_list.connect(self.ws_changed_list)
        self.worker.changed_queue.connect(self.ws_changed_queue)
        self.worker.tracker_state.connect(self.ws_tracker_state)
        self.worker.playing_show.connect(self.ws_changed_show)
        self.worker.prompt_for_update.connect(self.ws_prompt_update)
        self.worker.prompt_for_add.connect(self.ws_prompt_add)

        # Show main window
        if not (self.config['show_tray'] and self.config['start_in_tray']):
            self.show()

        # Start loading engine
        self.started = True
        self._busy(False)
        self.worker_call('start', self.r_engine_loaded, account)

    def reload(self, account=None, mediatype=None):
        if self.config['remember_columns']:
            self._store_columnstate()

        if account:
            self.account = account

            # Get API specific configuration
            self.api_config = self._get_api_config(account['api'])

        self.menu_columns.setEnabled(False)
        for action in self.menu_columns.actions():
            action.setChecked(
                action.text() in self.api_config['visible_columns'])
        self.menu_columns.setEnabled(True)

        self.show()
        self._busy(False)
        self.worker_call('reload', self.r_engine_loaded, account, mediatype)

    def closeEvent(self, event):
        if not self.started or not self.worker.engine.loaded:
            event.accept()
            if self.finish:
                QApplication.instance().quit()
        elif self.config['show_tray'] and self.config['close_to_tray']:
            event.ignore()
            self.s_hide()
        else:
            event.ignore()
            self._exit()

    def status(self, message):
        self.status_text.setText(message)
        print(message)

    def error(self, msg):
        self.status('Error: {}'.format(msg))
        QMessageBox.critical(self, 'Error', str(msg), QMessageBox.StandardButton.Ok)

    def fatal(self, msg):
        QMessageBox.critical(
            self, 'Fatal Error', "Fatal Error! Reason:\n\n{0}".format(msg), QMessageBox.StandardButton.Ok)
        self.accountman.set_default(None)
        self._busy()
        self.finish = False
        self.worker_call('unload', self.r_engine_unloaded)

    def worker_call(self, function, ret_function, *args, **kwargs):
        # Run worker in a thread
        self.worker.set_function(function, ret_function, *args, **kwargs)
        self.worker.start()

    # GUI Functions
    def _get_api_config(self, api):
        if self.config['columns_per_api']:
            if 'api' not in self.config:
                self.config['api'] = {}
            if api not in self.config['api']:
                self.config['api'][api] = dict(utils.qt_per_api_defaults)
            return self.config['api'][api]
        else:
            # API settings are universal
            return self.config

    def _save_config(self):
        utils.save_config(self.config, self.configfile)

    def _exit(self):
        self._busy()
        if self.config['remember_geometry']:
            self._store_geometry()
        if self.config['remember_columns']:
            self._store_columnstate()
        self.finish = True
        self.worker_call('unload', self.r_engine_unloaded)

    def _store_geometry(self):
        self.config['last_x'] = self.x()
        self.config['last_y'] = self.y()
        self.config['last_width'] = self.width()
        self.config['last_height'] = self.height()
        utils.save_config(self.config, self.configfile)

    def _store_columnstate(self):
        columns_state = {}

        state = self.view.horizontalHeader().saveState()
        columns_state = base64.b64encode(state).decode('ascii')

        self.api_config['columns_state'] = columns_state
        self._save_config()

    def _enable_widgets(self, enable):
        self.view.setEnabled(enable)
        self._enable_show_widgets(bool(self.selected_show_id and enable))

        self.action_add.setEnabled(enable)
        self.action_sync.setEnabled(enable)
        self.action_send.setEnabled(enable)
        self.action_retrieve.setEnabled(enable)
        self.action_reload.setEnabled(enable)

        self.show_filter.setEnabled(enable)
        self.show_filter_invert.setEnabled(enable)
        self.show_filter_casesens.setEnabled(enable)

    def _enable_show_widgets(self, enable):
        self.show_progress.setEnabled(enable)
        self.show_score.setEnabled(enable)
        self.show_progress_btn.setEnabled(enable)
        self.show_score_btn.setEnabled(enable)
        self.show_tags_btn.setEnabled(
            bool(self.mediainfo and self.mediainfo.get('can_tag') and enable))
        self.show_inc_btn.setEnabled(enable)
        self.show_dec_btn.setEnabled(enable)
        self.show_play_btn.setEnabled(enable)
        self.show_status.setEnabled(enable)
        self.action_play_next.setEnabled(enable)
        self.action_play_dialog.setEnabled(enable)
        self.action_altname.setEnabled(enable)
        self.action_delete.setEnabled(enable)
        self.action_details.setEnabled(enable)

    def _update_queue_counter(self, queue):
        self.queue_text.setText("Unsynced items: %d" % queue)

    def _update_tracker_info(self, status):
        state = status['state']
        timer = status['timer']
        paused = status['paused']

        if state == utils.Tracker.NOVIDEO:
            st = 'Listen'
        elif state == utils.Tracker.PLAYING:
            (m, s) = divmod(timer or 0, 60)
            st = "+{0}:{1:02d}{2}".format(m, s, ' [P]' if paused else '')
        elif state == utils.Tracker.UNRECOGNIZED:
            st = 'Unrecognized'
        elif state == utils.Tracker.NOT_FOUND:
            st = 'Not found'
        elif state == utils.Tracker.IGNORED:
            st = 'Ignored'
        else:
            st = '???'

        self.tracker_text.setText("Tracker: {}".format(st))

    def _update_config(self):
        self._apply_view()
        self._apply_tray()
        self._apply_filter_bar()
        # TODO: Reload listviews?

    def _apply_view(self):
        if self.config['inline_edit']:
            self.view.setEditTriggers(QAbstractItemView.EditTrigger.AllEditTriggers)
        else:
            self.view.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)

    def _apply_tray(self):
        if self.tray.isVisible() and not self.config['show_tray']:
            self.tray.hide()
        elif not self.tray.isVisible() and self.config['show_tray']:
            self.tray.show()
        if self.tray.isVisible():
            if self.config['tray_api_icon']:
                self.tray.setIcon(QtGui.QIcon(
                    utils.available_libs[self.account['api']][1]))
            else:
                self.tray.setIcon(self.windowIcon())

    def _apply_filter_bar(self):
        self.list_box.removeWidget(self.filter_bar_box)
        self.list_box.removeWidget(self.notebook)
        self.list_box.removeWidget(self.view)
        self.filter_bar_box.show()
        if self.config['filter_bar_position'] is FilterBar.PositionHidden:
            self.list_box.addWidget(self.notebook)
            self.list_box.addWidget(self.view)
            self.filter_bar_box.hide()
        elif self.config['filter_bar_position'] is FilterBar.PositionAboveLists:
            self.list_box.addWidget(self.filter_bar_box)
            self.list_box.addWidget(self.notebook)
            self.list_box.addWidget(self.view)
        elif self.config['filter_bar_position'] is FilterBar.PositionBelowLists:
            self.list_box.addWidget(self.notebook)
            self.list_box.addWidget(self.view)
            self.list_box.addWidget(self.filter_bar_box)

    def _busy(self, wait=False):
        if wait:
            self.busy_timer.start()
        else:
            self._enable_widgets(False)

    def _unbusy(self):
        if self.busy_timer.isActive():
            self.busy_timer.stop()
        else:
            self._enable_widgets(True)

    def _rebuild_statuses(self):
        # Rebuild statuses
        self.show_status.blockSignals(True)
        self.notebook.blockSignals(True)

        self.show_status.clear()

        # Clear notebook
        while self.notebook.count() > 0:
            self.notebook.removeTab(0)

        # Add one page per status
        for i, status in enumerate(self.mediainfo['statuses']):
            name = self.mediainfo['statuses_dict'][status]

            self.notebook.addTab(name)
            self.notebook.setTabData(i, status)

            self.show_status.addItem(name)

        self.notebook.addTab("All")

        self.show_status.blockSignals(False)
        self.notebook.blockSignals(False)

    def _recalculate_counts(self):
        showlist = self.worker.engine.get_list()

        self.counts = {status: 0 for status in self.mediainfo['statuses']}
        self.counts['!ALL'] = 0

        for show in showlist:
            self.counts[show['my_status']] += 1
            self.counts['!ALL'] += 1

        self._update_counts()

    def _update_counts(self):
        for page in range(self.notebook.count()):
            status = self.notebook.tabData(page)
            if status is not None:
                status_name = self.mediainfo['statuses_dict'][status]
            else:
                status_name = "All"
                status = "!ALL"

            self.notebook.setTabText(page, "{} ({})".format(
                status_name, self.counts[status]))

    def _rebuild_view(self):
        """
        Using a full showlist, rebuilds main view

        """
        showlist = self.worker.engine.get_list()
        altnames = self.worker.engine.altnames()
        library = self.worker.engine.library()

        # Set allowed ranges (this will be reported by the engine later)
        decimal_places = 0
        if isinstance(self.mediainfo['score_step'], float):
            decimal_places = len(
                str(self.mediainfo['score_step']).split('.')[1])

        self.show_score.setRange(0, self.mediainfo['score_max'])
        self.show_score.setDecimals(decimal_places)
        self.show_score.setSingleStep(self.mediainfo['score_step'])

        # Get the new list and pass it to our model
        self.view.setSortingEnabled(False)
        self.view.model().setFilterStatus(
            self.notebook.tabData(self.notebook.currentIndex()))
        self.view.model().sourceModel().setMediaInfo(self.mediainfo)
        self.view.model().sourceModel().setShowList(showlist, altnames, library)
        self.view.resizeRowsToContents()
        self.view.setSortingEnabled(True)

        self.s_filter_changed()

    def _init_view(self):
        # Set view options
        self.view.sortByColumn(self.config['sort_index'], QtCore.Qt.SortOrder(self.config['sort_order']))

        # Hide invisible columns
        for i, column in enumerate(self.view.model().sourceModel().columns):
            if column not in self.api_config['visible_columns']:
                self.view.setColumnHidden(i, True)

        self.view.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        self.view.horizontalHeader().setSectionResizeMode(3, QHeaderView.ResizeMode.Fixed)

        # Recover column state
        if self.config['remember_columns'] and isinstance(self.api_config['columns_state'], str):
            state = QtCore.QByteArray(base64.b64decode(
                self.api_config['columns_state']))
            self.view.horizontalHeader().restoreState(state)
        else:
            self.view.horizontalHeader().resizeSection(3, 70)
            self.view.horizontalHeader().resizeSection(4, 100)

    def _select_show(self, show):
        # Stop any running image timer
        if self.image_timer is not None:
            self.image_timer.stop()

        # Unselect show
        if not show:
            self.selected_show_id = None

            self.show_title.setText('Trackma-qt')
            self.show_image.setText('Trackma-qt')
            self.show_progress.setValue(0)
            self.show_score.setValue(0)
            self.show_progress_bar.setValue(0)
            self.show_progress_bar.setFormat('?/?')
            self._enable_show_widgets(False)

            return

        # Block signals
        self.show_status.blockSignals(True)

        # Set proper ranges
        if show['total']:
            self.show_progress.setMaximum(show['total'])
            self.show_progress_bar.setFormat('%v/%m')
            self.show_progress_bar.setMaximum(show['total'])
            # Regenerate Play Episode Menu
            self.generate_episode_menus(
                self.menu_play, show['total'], show['my_progress'])
        else:
            self.show_progress.setMaximum(
                utils.estimate_aired_episodes(show) or 10000)
            self.generate_episode_menus(
                self.menu_play, utils.estimate_aired_episodes(show), show['my_progress'])
            self.show_progress_bar.setFormat(
                '{}/?'.format(show['my_progress']))

        # Update information
        metrics = QtGui.QFontMetrics(self.show_title.font())
        title = metrics.elidedText(
            show['title'], QtCore.Qt.TextElideMode.ElideRight, self.show_title.width())
        self.show_title.setText(title)

        self.show_progress.setValue(show['my_progress'])
        self.show_status.setCurrentIndex(
            self.mediainfo['statuses'].index(show['my_status']))
        self.show_score.setValue(show['my_score'])

        # Enable relevant buttons
        self._enable_show_widgets(True)

        # Download image or use cache
        if show.get('image_thumb') or show.get('image'):
            if self.image_worker is not None:
                self.image_worker.cancel()

            utils.make_dir(utils.to_cache_path())
            filename = utils.to_cache_path("%s_%s_%s.jpg" % (
                self.api_info['shortname'], self.api_info['mediatype'], show['id']))

            if os.path.isfile(filename):
                self.s_show_image(filename)
            else:
                if "imaging_available" in os.environ:
                    self.show_image.setText('Waiting...')
                    self.image_timer.start()
                else:
                    self.show_image.setText('Not available')
        else:
            self.show_image.setText('No image')

        if show['total']:
            self.show_progress_bar.setValue(show['my_progress'])
        else:
            self.show_progress_bar.setValue(0)

        # Make it global
        self.selected_show_id = show['id']

        # Unblock signals
        self.show_status.blockSignals(False)

    def generate_episode_menus(self, menu, max_eps=1, watched_eps=0):
        bp_top = 5  # No more than this many submenus/episodes in the root menu
        bp_mid = 10  # No more than this many submenus in submenus
        bp_btm = 13  # No more than this many episodes in the submenus
        # The number of episodes where we ditch the submenus entirely since Qt doesn't deserve this abuse
        breakpoint_no_menus = bp_top * bp_btm * bp_mid * bp_mid

        menu.clear()
        # Make basic actions
        action_play_next = QAction(
            getIcon('media-skip-forward'), 'Play &Next Episode', self)
        action_play_next.triggered.connect(lambda: self.s_play(True))
        action_play_last = QAction(
            getIcon('view-refresh'), 'Play Last Watched Ep (#%d)' % watched_eps, self)
        action_play_last.triggered.connect(lambda: self.s_play(False))
        action_play_dialog = QAction('Play Episode...', self)
        action_play_dialog.setStatusTip('Select an episode to play.')
        action_play_dialog.triggered.connect(self.s_play_number)

        menu.addAction(action_play_next)
        menu.addAction(action_play_last)

        if max_eps < 1 or max_eps > breakpoint_no_menus:
            menu.addAction(action_play_dialog)
            return menu
        menu.addSeparator()

        ep_actions = []
        for ep in range(1, max_eps+1):
            action = QAction('Ep. %d' % ep, self)
            action.triggered.connect(self.s_play_ep_number(action, ep))
            if ep <= watched_eps:
                action.setIcon(self.ep_icons['all'])
            else:
                action.setIcon(self.ep_icons['none'])
            ep_actions.append(action)

        if max_eps <= bp_top:
            # Just put the eps in the root menu
            for action in ep_actions:
                menu.addAction(action)

        else:
            # We need to go deeper. For now, put all the episodes into bottom-level submenus.
            # I don't like this scoping. If you find a way to transfer ownership of the submenu to the menu feel free to fix this.
            self.play_ep_submenus = []
            # A bit hacky but avoids a special case for the first submenu
            current_actions = bp_btm + 1
            for action in ep_actions:
                if current_actions >= bp_btm:
                    current_actions = 0
                    length = len(self.play_ep_submenus)
                    self.play_ep_submenus.append(
                        QMenu('Episodes %d-%d:' % (length*bp_btm + 1, min((length+1)*bp_btm, max_eps)), menu))
                    if watched_eps > min((length+1)*bp_btm, max_eps):
                        self.play_ep_submenus[-1].setIcon(self.ep_icons['all'])
                    elif watched_eps > length*bp_btm:
                        self.play_ep_submenus[-1].setIcon(
                            self.ep_icons['part'])
                    else:
                        self.play_ep_submenus[-1].setIcon(
                            self.ep_icons['none'])
                self.play_ep_submenus[-1].addAction(action)
                current_actions += 1

            # Now to put the bottom level menus into other things
            if len(self.play_ep_submenus) <= bp_top:  # Straight into the root menu, easy!
                for submenu in self.play_ep_submenus:
                    menu.addMenu(submenu)
            else:  # For now, put them into another level of submenus
                self.play_ep_sub2menus = []
                current_menus = bp_mid + 1
                for s in self.play_ep_submenus:
                    if current_menus >= bp_mid:
                        current_menus = 0
                        length = len(self.play_ep_sub2menus)
                        self.play_ep_sub2menus.append(QMenu(
                            'Episodes %d-%d:' % (length*bp_btm*bp_mid + 1, min((length+1)*bp_btm*bp_mid, max_eps)), s))
                    self.play_ep_sub2menus[-1].addMenu(s)
                    if watched_eps > min((length+1)*bp_btm*bp_mid, max_eps):
                        self.play_ep_sub2menus[-1].setIcon(
                            self.ep_icons['all'])
                    elif watched_eps > length*bp_btm*bp_mid:
                        self.play_ep_sub2menus[-1].setIcon(
                            self.ep_icons['part'])
                    else:
                        self.play_ep_sub2menus[-1].setIcon(
                            self.ep_icons['none'])
                    current_menus += 1

                if len(self.play_ep_sub2menus) <= bp_top:
                    for submenu in self.play_ep_sub2menus:
                        menu.addMenu(submenu)
                else:
                    # I seriously hope this additional level is not needed, but maybe someone will want to set smaller breakpoints.
                    self.play_ep_sub3menus = []
                    current_menus = bp_mid + 1
                    for s in self.play_ep_sub2menus:
                        if current_menus >= bp_mid:
                            current_menus = 0
                            length = len(self.play_ep_sub3menus)
                            self.play_ep_sub3menus.append(QMenu(
                                'Episodes %d-%d:' % (length*bp_btm*bp_mid*bp_mid + 1, min((length+1)*bp_btm*bp_mid*bp_mid, max_eps)), s))
                        self.play_ep_sub3menus[-1].addMenu(s)
                        if watched_eps > min((length+1)*bp_btm*bp_mid*bp_mid, max_eps):
                            self.play_ep_sub3menus[-1].setIcon(
                                self.ep_icons['all'])
                        elif watched_eps > length*bp_btm*bp_mid*bp_mid:
                            self.play_ep_sub3menus[-1].setIcon(
                                self.ep_icons['part'])
                        else:
                            self.play_ep_sub3menus[-1].setIcon(
                                self.ep_icons['none'])
                        current_menus += 1
                    # No more levels, our sanity check earlier ensured that.
                    for submenu in self.play_ep_sub3menus:
                        menu.addMenu(submenu)
        return menu

    # Slots
    def s_hide(self):
        if self.isVisible():
            self.was_maximized = self.isMaximized()
            self.hide()
        else:
            self.setGeometry(self.geometry())
            if self.was_maximized:
                self.showMaximized()
            else:
                self.show()

    def s_tray_clicked(self, reason):
        if reason == QSystemTrayIcon.ActivationReason.Trigger:
            self.s_hide()

    def s_busy(self):
        self._enable_widgets(False)

    def s_show_selected(self, new, old=None):
        if new:
            index = self.view.model().index(new.row(), 0)
            selected_id = self.view.model().data(index)

            if selected_id:
                show = self.worker.engine.get_show_info(selected_id)
                self._select_show(show)
        else:
            self._select_show(None)

    def s_update_sort(self, index, order):
        self.config['sort_index'] = index
        self.config['sort_order'] = order.value

    def s_download_image(self):
        show = self.worker.engine.get_show_info(self.selected_show_id)
        self.show_image.setText('Downloading...')
        filename = utils.to_cache_path("%s_%s_%s.jpg" % (
            self.api_info['shortname'], self.api_info['mediatype'], show['id']))

        self.image_worker = ImageWorker(
            show.get('image_thumb') or show['image'], filename, (100, 140))
        self.image_worker.finished.connect(self.s_show_image)
        self.image_worker.start()

    def s_tab_changed(self, index):
        # Change the filter of the main view to the specified status
        status = self.notebook.tabData(index)
        self.view.model().setFilterStatus(status)
        self.view.resizeRowsToContents()  # TODOMVC : Find a faster way

        self.s_show_selected(None)
        self.s_filter_changed()  # Refresh filter

    def s_filter_changed(self):
        # TODOMVC DEPRECATED
        expression = self.show_filter.text()

        # Determine if a show matches a filter. True -> match -> do not hide
        # Advanced search: Separate the expression into specific field terms, fail if any are not met
        if ':' in expression:
            exprs = expression.split(' ')
            expr_dict = {}
            expr_list = []
            for expr in exprs:
                if ':' in expr:
                    expr_terms = expr.split(':', 1)
                    if expr_terms[0] in self.column_keys:
                        col = self.column_keys[expr_terms[0]]
                        sub_expr = expr_terms[1].replace(
                            '_', ' ').replace('+', ' ')
                        expr_dict[col] = sub_expr
                    else:  # If it's not a field key, let it be a regular search term
                        expr_list.append(expr)
                else:
                    expr_list.append(expr)
            expression = ' '.join(expr_list)
            self.view.model().setFilterColumns(expr_dict)

        if self.show_filter_casesens.isChecked():
            self.view.model().setFilterCaseSensitivity(QtCore.Qt.CaseSensitivity.CaseSensitive)
        else:
            self.view.model().setFilterCaseSensitivity(QtCore.Qt.CaseSensitivity.CaseInsensitive)
        self.view.model().setFilterFixedString(expression)

    def s_plus_episode(self):
        self._busy(True)
        self.worker_call('set_episode', self.r_generic,
                         self.selected_show_id, self.show_progress.value()+1)

    def s_rem_episode(self):
        if not self.show_progress.value() <= 0:
            self._busy(True)
            self.worker_call('set_episode', self.r_generic,
                             self.selected_show_id, self.show_progress.value()-1)

    def s_set_episode(self, showid=None, ep=None):
        self._busy(True)
        self.worker_call('set_episode', self.r_generic,
                         showid or self.selected_show_id, ep if ep is not None else self.show_progress.value())

    def s_set_score(self, showid=None, score=None):
        self._busy(True)

        if not showid:
            showid = self.selected_show_id
        if score is None:
            score = round(self.show_score.value(), self.show_score.decimals())

        self.worker_call('set_score', self.r_generic, showid, score)

    def s_set_status(self, index):
        if self.selected_show_id:
            self._busy(True)
            self.worker_call('set_status', self.r_generic,
                             self.selected_show_id, self.mediainfo['statuses'][index])

    def s_set_tags(self):
        show = self.worker.engine.get_show_info(self.selected_show_id)
        if 'my_tags' in show and show['my_tags']:
            tags = show['my_tags']
        else:
            tags = ''
        tags, ok = QInputDialog.getText(self, 'Edit Tags',
                                        'Enter desired tags (comma separated)',
                                        text=tags)
        if ok:
            self.s_edit_tags(show, tags)

    def s_edit_tags(self, show, tags):
        self._busy(True)
        self.worker_call('set_tags', self.r_generic, show['id'], tags)

    def s_play(self, play_next, episode=0):
        if self.selected_show_id:
            show = self.worker.engine.get_show_info(self.selected_show_id)

            # episode = 0 # Engine plays next unwatched episode
            if not play_next and not episode:
                episode = self.show_progress.value()

            self._busy(True)
            self.worker_call('play_episode', self.r_play_episode, show, episode)

    def s_play_random(self):
        self._busy(True)
        self.worker_call('play_random', self.r_play_episode)

    def s_play_number(self):
        if self.selected_show_id:
            show = self.worker.engine.get_show_info(self.selected_show_id)
            ep_default = 1
            ep_min = 1
            ep_max = utils.estimate_aired_episodes(show)
            if not ep_max:
                # If we don't know the total just allow anything
                ep_max = show['total'] or 10000

            episode, ok = QInputDialog.getInt(self, 'Play Episode',
                                              'Enter an episode number of %s to play:' % show['title'],
                                              ep_default, ep_min, ep_max)

            if ok:
                self.s_play(False, episode)

    def s_play_ep_number(self, action, number):
        return lambda: [action.setIcon(self.ep_icons['part']), self.s_play(False, number)]

    def s_delete(self):
        if self.selected_show_id:
            show = self.worker.engine.get_show_info(self.selected_show_id)
            reply = QMessageBox.question(self, 'Confirmation',
                                         'Are you sure you want to delete %s?' % show['title'],
                                         QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

            if reply == QMessageBox.StandardButton.Yes:
                self.worker_call('delete_show', self.r_generic, show)

    def s_scan_library(self):
        self.worker_call('scan_library', self.r_library_scanned, rescan=False)

    def s_rescan_library(self):
        self.worker_call('scan_library', self.r_library_scanned, rescan=True)

    def s_altname(self):
        show = self.worker.engine.get_show_info(self.selected_show_id)
        current_altname = self.worker.engine.altname(self.selected_show_id)

        new_altname, ok = QInputDialog.getText(self, 'Alternative title',
                                               'Set the new alternative title for %s (blank to remove):' % show[
                                                   'title'],
                                               text=current_altname)

        if ok:
            self.worker.engine.altname(self.selected_show_id, str(new_altname))
            self.ws_changed_show(show, altname=new_altname)

    def s_open_folder(self):
        try:
            self.worker.engine.open_show_folder(self.selected_show_id)
        except utils.EngineError as e:
            self.error(e.args[0])

    def s_retrieve(self):
        queue = self.worker.engine.get_queue()

        if queue:
            reply = QMessageBox.question(self, 'Confirmation',
                                         'There are %d unsynced changes. Do you want to send them first? (Choosing No will discard them!)' % len(
                                             queue),
                                         QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Cancel)

            if reply == QMessageBox.StandardButton.Yes:
                self.s_send(True)
            elif reply == QMessageBox.StandardButton.No:
                self._busy(True)
                self.worker_call('list_download', self.r_list_retrieved)
        else:
            self._busy(True)
            self.worker_call('list_download', self.r_list_retrieved)

    def s_send(self, retrieve=False):
        self._busy(True)
        if retrieve:
            self.worker_call('list_upload', self.s_retrieve)
        else:
            self.worker_call('list_upload', self.r_generic_ready)

    def s_switch_account(self):
        if not self.accountman_widget:
            self.accountman_create()
        else:
            self.accountman_widget.rebuild()

        self.accountman_widget.setModal(True)
        self.accountman_widget.show()

    def s_show_image(self, filename):
        self.show_image.setPixmap(QtGui.QPixmap(filename))

    def s_show_details(self):
        if not self.selected_show_id:
            return

        show = self.worker.engine.get_show_info(self.selected_show_id)

        self.detailswindow = DetailsDialog(None, self.worker, show)
        self.detailswindow.setModal(True)
        self.detailswindow.show()

    def s_add(self):
        current_status = self.notebook.tabData(self.notebook.currentIndex())

        self.addwindow = AddDialog(None, self.worker, current_status)
        self.addwindow.setModal(True)
        self.addwindow.show()

    def s_mediatype(self, action):
        index = action.data()
        if index is not None:
            mediatype = self.api_info['supported_mediatypes'][index]
            self.reload(None, mediatype)

    def s_settings(self):
        dialog = SettingsDialog(
            None, self.worker, self.config, self.configfile)
        dialog.saved.connect(self._update_config)
        dialog.exec()

    def s_about(self):
        QMessageBox.about(self, 'About Trackma-qt %s' % utils.VERSION,
                          '<p><b>About Trackma-qt %s</b></p><p>Trackma is an open source client for media tracking websites.</p>'
                          '<p>This program is licensed under the GPLv3, for more information read COPYING file.</p>'
                          '<p>Thanks to all contributors. To see all contributors see AUTHORS file.</p>'
                          '<p>Copyright (C) z411 - Icon by shuuichi</p>'
                          '<p><a href="https://github.com/z411/trackma">https://github.com/z411/trackma</a></p>' % utils.VERSION)

    def s_about_qt(self):
        QMessageBox.aboutQt(self, 'About Qt')

    def s_show_menu_columns(self, pos):
        globalPos = self.sender().mapToGlobal(pos)
        globalPos += QtCore.QPoint(3, 3)
        self.menu_columns.exec(globalPos)

    def s_toggle_column(self, visible):
        w = self.sender()
        index, column_name = w.data(), w.text()
        MIN_WIDTH = 30  # Width to restore columns to if too small to see

        if visible:
            if column_name not in self.api_config['visible_columns']:
                self.api_config['visible_columns'].append(str(column_name))
        else:
            if column_name in self.api_config['visible_columns']:
                self.api_config['visible_columns'].remove(column_name)

        self._save_config()

        self.view.setColumnHidden(index, not visible)
        if visible and self.view.columnWidth(index) < MIN_WIDTH:
            self.view.setColumnWidth(index, MIN_WIDTH)

    # Worker slots
    def ws_changed_status(self, classname, msgtype, msg):
        if msgtype != messenger.TYPE_DEBUG:
            self.status('{}: {}'.format(classname, msg))
        elif self.debug:
            print('[D] {}: {}'.format(classname, msg))

    def ws_changed_show(self, show, is_playing=False, episode=None, altname=None):
        if show:
            if not self.view:
                return  # List not built yet; can be safely avoided

            # Update the view of the updated show
            self.view.model().sourceModel().update(show['id'], is_playing)

            if show['id'] == self.selected_show_id:
                self._select_show(show)

            if is_playing and self.config['show_tray'] and self.config['notifications']:
                if episode == (show['my_progress'] + 1):
                    delay = self.worker.engine.get_config(
                        'tracker_update_wait_s')
                    self.tray.showMessage('Trackma Tracker', "Playing %s %s. Will update in %d seconds." % (
                        show['title'], episode, delay))

    def ws_changed_show_status(self, show, old_status=None):
        # Update the view of the new show
        self.view.model().sourceModel().update(show['id'])

        # Update counts
        self.counts[show['my_status']] += 1
        self.counts[old_status] -= 1
        self._update_counts()

        # Set notebook to the new page
        self.notebook.setCurrentIndex(
            self.mediainfo['statuses'].index(show['my_status']))
        # Refresh filter
        self.s_filter_changed()

    def ws_changed_list(self, show):
        self._rebuild_view()
        self._recalculate_counts()
        self.s_filter_changed()

    def ws_changed_queue(self, queue):
        self._update_queue_counter(queue)

    def ws_tracker_state(self, status):
        self._update_tracker_info(status)

    def ws_prompt_update(self, show, episode):
        box = QMessageBox(self)
        box.setWindowTitle("Update prompt")
        box.setText(f"Do you want to update {show['title']} to {episode}?")
        box.setIcon(QMessageBox.Icon.Question)
        box.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        box.setAttribute(QtCore.Qt.WidgetAttribute.WA_ShowWithoutActivating)
        box.setModal(False)
        box.accepted.connect(lambda:
                self.worker_call('set_episode', self.r_generic,
                show['id'], episode))
        box.show()

    def ws_prompt_add(self, show, episode):
        page = self.notebook.currentIndex()
        current_status = self.mediainfo['statuses'][page]

        addwindow = AddDialog(
            None, self.worker, current_status, default=show['title'])
        addwindow.setModal(True)
        if addwindow.exec():
            self.worker_call('set_episode', self.r_generic,
                             addwindow.selected_show['id'], episode)

    # Responses from the engine thread
    def r_generic(self):
        self._unbusy()

    def r_generic_ready(self):
        self._unbusy()
        self.status('Ready.')

    def r_engine_loaded(self, result):
        if result['success']:
            showlist = self.worker.engine.get_list()
            altnames = self.worker.engine.altnames()
            library = self.worker.engine.library()

            # Set globals
            self.api_info = self.worker.engine.api_info
            self.mediainfo = self.worker.engine.mediainfo

            # Rebuild statuses
            self._rebuild_statuses()

            # Build mediatype menu
            for action in self.mediatype_actiongroup.actions():
                self.mediatype_actiongroup.removeAction(action)

            for n, mediatype in enumerate(self.api_info['supported_mediatypes']):
                action = QAction(mediatype, self, checkable=True)
                if mediatype == self.api_info['mediatype']:
                    action.setChecked(True)
                else:
                    action.setData(n)
                self.mediatype_actiongroup.addAction(action)
                self.menu_mediatype.addAction(action)

            # Show API info
            self.api_icon.setPixmap(QtGui.QPixmap(
                utils.available_libs[self.account['api']][1]))
            if self.config['tray_api_icon']:
                self.tray.setIcon(QtGui.QIcon(
                    utils.available_libs[self.account['api']][1]))
            self.api_user.setText(
                self.worker.engine.get_userconfig('username'))
            self.setWindowTitle("Trackma-qt %s [%s (%s)]" % (
                utils.VERSION, self.api_info['name'], self.api_info['mediatype']))

            # Show tracker info
            tracker_info = self.worker.engine.tracker_status()
            if tracker_info:
                self._update_tracker_info(tracker_info)

            # Build our main view and show total counts
            self._rebuild_view()
            self._init_view()
            self._recalculate_counts()

            self.s_show_selected(None)

            self.status('Ready.')

        self._unbusy()

    def r_list_retrieved(self, result):
        if result['success']:
            self._rebuild_view()
            self._recalculate_counts()

            self.status('Ready.')

        self._unbusy()

    def r_library_scanned(self, result):
        if result['success']:
            self._rebuild_view()

            self.status('Ready.')

        self._unbusy()

    def r_engine_unloaded(self, result):
        if result['success']:
            self.close()
            if not self.finish:
                self.s_switch_account()

    def r_play_episode(self, result):
        if result['success']:
            args = result['result']
            if len(args) > 1:
                # QtCore.QProcess.startDetached(args[0], args[1:])
                process = QtCore.QProcess()
                for attr in ['setStandardErrorFile', 'setStandardOutputFile']:
                    getattr(process, attr)(QtCore.QProcess.nullDevice())
                process.setProgram(args[0])
                process.setArguments(args[1:])
                process.startDetached()

        self._unbusy()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import traceback

TYPE_DEBUG = 1
TYPE_INFO = 2
# TYPE_ERROR = 3
# TYPE_FATAL = 4
TYPE_WARN = 5


class Messenger:
    _handler = None

    def __init__(self, handler, classname):
        self._handler = handler
        self.classname = classname

    def set_handler(self, handler):
        self._handler = handler

    def with_classname(self, classname):
        return Messenger(self._handler, classname)

    def _call_handler(self, msgs, msg_type):
        if self._handler:
            cn, msg = self._parse_msgs(msgs)
            self._handler(cn, msg_type, msg)

    def _parse_msgs(self, msgs):
        if len(msgs) >= 2:
            return (msgs[0], " ".join(msgs[1:]) if msgs[2:] else msgs[1])
        return (self.classname, msgs[0])

    def debug(self, *msgs):
        self._call_handler(msgs, TYPE_DEBUG)

    def info(self, *msgs):
        self._call_handler(msgs, TYPE_INFO)

    def warn(self, *msgs):
        self._call_handler(msgs, TYPE_WARN)

    def exception(self, *msgs):
        if not self._handler:
            return
        cn, exc_info = self._parse_msgs(msgs[:2])
        exc_type, exc_value, exc_traceback = exc_info
        for block in traceback.format_exception(exc_type, exc_value, exc_traceback):
            for line in block.splitlines():
                self._handler(cn, TYPE_DEBUG, line)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt6 import QtCore, QtGui
from PyQt6.QtWidgets import QDialog, QDialogButtonBox, QGridLayout, QPushButton, QVBoxLayout


class ThemedColorPicker(QDialog):
    def __init__(self, parent=None, default=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle('Select Color')
        layout = QVBoxLayout()
        colorbox = QGridLayout()
        self.colorString = default

        self.groups = [0, 1, 2]
        self.roles = [1, 2, 3, 4, 5, 11, 12, 16]  # Only use background roles
        self.colors = []
        row = 0
        # Make colored buttons for selection
        for group in self.groups:
            col = 0
            for role in self.roles:
                self.colors.append(QPushButton())
                self.colors[-1].setStyleSheet('background-color: ' + QtGui.QColor(
                    QtGui.QPalette().color(QtGui.QPalette.ColorGroup(group), QtGui.QPalette.ColorRole(role))).name())
                self.colors[-1].setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.colors[-1].clicked.connect(self.s_select(group, role))
                colorbox.addWidget(self.colors[-1], row, col, 1, 1)
                col += 1
            row += 1
        bottombox = QDialogButtonBox()
        bottombox.addButton(QDialogButtonBox.StandardButton.Ok)
        bottombox.addButton(QDialogButtonBox.StandardButton.Cancel)
        bottombox.accepted.connect(self.accept)
        bottombox.rejected.connect(self.reject)
        layout.addLayout(colorbox)
        layout.addWidget(bottombox)
        self.setLayout(layout)

    def s_select(self, group, role):
        return lambda: self.select(group, role)

    def select(self, group, role):
        self.colorString = str(group) + ',' + str(role)

    @staticmethod
    def do(parent=None, default=None):
        dialog = ThemedColorPicker(parent, default)
        result = dialog.exec()

        if result == QDialog.DialogCode.Accepted:
            return dialog.colorString
        else:
            return None



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

gtk_dir = os.path.dirname(__file__)


def main():
    import signal
    import sys
    from trackma import utils
    from .application import TrackmaApplication

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    print("Trackma GTK v{}".format(utils.VERSION))
    app = TrackmaApplication()
    sys.exit(app.run(sys.argv))
